% Note: The code is optimized assuming that documents with more words come
% first. If this is not the case, the parallelization may not be optimal.

addpath('../common');

% set the random number generation seed for reproducible data
RandStream.setGlobalStream(RandStream('mt19937ar','Seed', 1));

% generate data. data should be a 1D struct array with one field called
% "words", where each element in the struct array is a document.
% data(d).words should be the words in document d, represented as indices.
% That is, data(d).words should be a 1D array with size the number of words
% in document d. Each word should be an index into the dictionary, where
% the minimum index starts at 0. If data(3).words = [0 3 1 3], it means
% that document 3 has the 0th word, 3rd word, 1st word, and 3rd word.
%
% dictSize is the size of the dictionary. words should be in
% [0, dictSize-1]
[data, dictSize] = generate_bar_data();

% This is a flag to visualize things specific to the bars dataset.
isBarsDataset = true;

% Run the sampler
initialClusters = 1;
dispOn = true;
numProcessors = 8;
gamma = 1;
alpha = 1;
endtime = 100;
% The number of calls to the MEX function. Each MEX function runs 20
% iterations. numits=25 will do 25*20 iterations.
numits = 25;

% Run the sampling algorithm.
model = run_HDP_subclusters(data, dictSize, initialClusters, dispOn,  ...
    numProcessors, gamma, alpha, endtime, numits, isBarsDataset);



% Visualize some of the information contained in the model struct. These
% are probably fairly useless to see, but it gives you an idea on how to
% access the sampled data.
figure(3);
pos = get(sfigure(3),'Position');
pos(3) = 550;
pos(4) = 650;
set(sfigure(3),'Position', pos);

% Plot statistics of the corpus
subplot(4,1,1);
num_topics = numel(model.dishes);
topic_indices = zeros(num_topics, 1);
topic_betas = zeros(num_topics, 1);
for k=1:num_topics
    topic_indices(k) = model.dishes(k).dishInd;
    topic_betas(k) = model.dishes(k).beta;
end
stem(topic_indices, topic_betas);
title('Global Topic Probabilities');
xlabel('Topic Index, k');
ylabel('beta_k');
    
% Get statistics of each document
num_docs = numel(model.restaurants);
labels = cell(num_docs, 1);
for d=1:num_docs
    labels{d} = model.restaurants(d).customers;
end

% Plot the statistics for documents 1:3
for d=1:3
    subplot(4,1,d+1);
    counts = hist(labels{d}, 0:max(topic_indices));
    stem(0:max(topic_indices), counts);
    title(['Histogram of Topic Assignments for Document ' num2str(d)]);
    xlabel('Topic Index, k');
    ylabel('# of words in topic');
end