#ifndef _CORPUS_H_INCLUDED_
#define _CORPUS_H_INCLUDED_


class corpus
{
private:
   int N; // number of documents
   arr(arr(unsigned int)) words;
   arr(unsigned int) lengths;
   int HOW; // number of held out words per doc

public:
   corpus();
   corpus(const mxArray* data, int _HOW);
   corpus(const corpus& that);
   corpus& operator=(const corpus& that);
   void copy(const corpus& that);
   virtual ~corpus();

   void setHOW(int _HOW);

   unsigned int* getDocument(int d);
   unsigned int getDocumentLength(int d);
   unsigned int* getDocumentHO(int d);
   unsigned int getDocumentLengthHO(int d);
};





inline corpus::corpus() : N(0), words(NULL), lengths(NULL)
{
}

inline corpus::corpus(const mxArray* data, int _HOW) : HOW(_HOW)
{
   checkInput(data, STRUCT);
   N = mxGetNumberOfElements(data);

   words = allocate_memory< arr(unsigned int) >(N);
   lengths = allocate_memory< unsigned int >(N);

   for (int i=0; i<N; i++)
   {
      words[i] = getArrayInput<unsigned int>(getField(data,i,"words"));
      lengths[i] = mxGetNumberOfElements(getField(data,i,"words"));
   }
}

inline corpus::corpus(const corpus& that)
{
   copy(that);
}

inline corpus& corpus::operator=(const corpus& that)
{
   if (this!=&that)
   {
      if (words!=NULL) deallocate_memory(words);
      if (lengths!=NULL) deallocate_memory(lengths);
      copy(that);
   }
   return *this;
}
inline void corpus::copy(const corpus& that)
{ 
   N = that.N;
   words = allocate_memory< arr(unsigned int) >(N);
   lengths = allocate_memory< unsigned int >(N);
   copy_memory(words, that.words, sizeof(unsigned int*)*N);
   copy_memory(lengths, that.lengths, sizeof(unsigned int)*N);
}

inline corpus::~corpus()
{
   if (words!=NULL) deallocate_memory(words);
   if (lengths!=NULL) deallocate_memory(lengths);
}

inline void corpus::setHOW(int _HOW)
{
   HOW = _HOW;
}

inline unsigned int* corpus::getDocument(int d)
{
   return words[d] + HOW;
}
inline unsigned int corpus::getDocumentLength(int d)
{
   long length = (long)(lengths[d]) - HOW;
   if (length<0)
      mexErrMsgTxt("Too many held out words!\n");
   return length;
}


inline unsigned int* corpus::getDocumentHO(int d)
{
   return words[d];
}
inline unsigned int corpus::getDocumentLengthHO(int d)
{
   if (HOW > lengths[d])
      mexErrMsgTxt("Too many held out words!\n");
   return HOW;
}
#endif