

#include "hdp_subclusters.h"
#include "categorical.h"
#include "tableFuncs.h"
#include "myfuncs.h"
#include "sample_categorical.h"

#ifndef pi
#define pi 3.14159265
#endif


typedef unordered_map<int,dish>::iterator dishIterator;
typedef unordered_map<int,restaurantDish>::iterator rdishIterator;
typedef unordered_map<int,restaurantDishComps>::iterator rdishCompsIterator;

// --------------------------------------------------------------------------
// -- hdp_subclusters
// --   copy constructor;
// --------------------------------------------------------------------------
hdp_subclusters::hdp_subclusters(const hdp_subclusters& that)
{
   copy(that);
}
// --------------------------------------------------------------------------
// -- operator=
// --   assignment operator
// --------------------------------------------------------------------------
hdp_subclusters& hdp_subclusters::operator=(const hdp_subclusters& that)
{
   if (this != &that)
   {
   	cleanup();
      copy(that);
   }
   return *this;
}
// --------------------------------------------------------------------------
// -- copy
// --   returns a copy of this
// --------------------------------------------------------------------------
void hdp_subclusters::copy(const hdp_subclusters& that)
{
	NThreads = that.NThreads;
   D = that.D;
   K = that.K;
   documents = that.documents;

   gamma = that.gamma;
   alpha = that.alpha;
   loggamma = fastlog(gamma);
   logalpha = fastlog(alpha);
   diralpha = that.diralpha;

   customers = that.customers;
   randCustomers = that.randCustomers;
   restaurants = that.restaurants;
   dishes = that.dishes;
   hyper = that.hyper;

   k2zDishPtrs.resize(max(500,max(D,hyper.getD())));
   z2k.resize(max(500,max(D,hyper.getD())));

   probsArr = allocate_memory< arr(Real) >(NThreads);
   probsArrIndices = allocate_memory< arr(int) >(NThreads);
   rArr = allocate_memory< gsl_rng* >(NThreads);

   for (int t=0; t<NThreads; t++)
   {
   	probsArr[t] = allocate_memory<Real>(max(500,max(D,hyper.getD())));
   	probsArrIndices[t] = allocate_memory<int>(max(500,max(D,hyper.getD())));
   	rArr[t] = initialize_gsl_rand(rand());
   }
}

// --------------------------------------------------------------------------
// -- hdp_subclusters
// --   constructor; intializes to all the values given
// --------------------------------------------------------------------------
hdp_subclusters::hdp_subclusters(const mxArray* params, const mxArray* _data) :
	documents(_data,0)
{
	D = mxGetNumberOfElements(_data);
   int W = getInput<double>(getField(params,0,"W"));

   diralpha = getInput<double>(getField(params,0,"diralpha"));
   NThreads = getInput<double>(getField(params,0,"NThreads"));
   int HOW = getInput<double>(getField(params,0,"HOW"));
   documents.setHOW(HOW);
   logstirling = getArrayInput<double>(getField(params,0,"logstirling"));
   logstirlingW = mxGetM(getField(params,0,"logstirling"));
   bestNumTables = getArrayInput<int>(getField(params,0,"bestNumTables"));
   bestNumTablesN = mxGetM(getField(params,0,"bestNumTables"));

   always_splittable = getInput<bool>(getField(params,0,"always_splittable"));
   hyper = dirichlet(NThreads, W, diralpha);


   k2zDishPtrs.resize(max(500,max(D,hyper.getD())));
   z2k.resize(max(500,max(D,hyper.getD())));

	probsArr = allocate_memory< arr(Real) >(NThreads);
   probsArrIndices = allocate_memory< arr(int) >(NThreads);
   rArr = allocate_memory< gsl_rng* >(NThreads);
   for (int t=0; t<NThreads; t++)
   {
   	probsArr[t] = allocate_memory<Real>(max(500,max(D,hyper.getD())));
   	probsArrIndices[t] = allocate_memory<int>(max(500,max(D,hyper.getD())));
   	rArr[t] = initialize_gsl_rand(rand());
   }

   // set the right vector sizes
   customers.resize(D);
   tcustomers.resize(D);
   randCustomers.resize(D);
   restaurants.resize(D);

  	for (int i=0; i<D; i++)
  	{
  		int W = documents.getDocumentLength(i);
  		customers[i].resize(W);
  		tcustomers[i].resize(W);
  		randCustomers[i].resize(W);
  		if (W>=logstirlingW)
  			mexErrMsgTxt("LogStirlingTable not large enough\n");
  		if (W>= bestNumTablesN)
  			mexErrMsgTxt("BestNumTables  not large enough\n");
  	}
}


// --------------------------------------------------------------------------
// -- ~hdp_subclusters
// --   destructor
// --------------------------------------------------------------------------
hdp_subclusters::~hdp_subclusters()
{
	cleanup();
}

void hdp_subclusters::cleanup()
{
	for (int t=0; t<NThreads; t++)
	{
		deallocate_memory(probsArr[t]);
		deallocate_memory(probsArrIndices[t]);
		gsl_rng_free(rArr[t]);
	}
	deallocate_memory(probsArr);
	deallocate_memory(probsArrIndices);
	deallocate_memory(rArr);
}


// --------------------------------------------------------------------------
// -- initialize
// --   populates the initial statistics
// --------------------------------------------------------------------------
void hdp_subclusters::initialize(const mxArray* inputs)
{
	gamma = getInput<double>(getField(inputs,0,"gamma"));
   alpha = getInput<double>(getField(inputs,0,"alpha"));
	loggamma = fastlog(gamma);
   logalpha = fastlog(alpha);


	const mxArray* dishesIn = getField(inputs,0,"dishes");
	K = mxGetNumberOfElements(dishesIn);

	emptyBeta = 1;
	for (int k=0; k<K; k++)
	{
		int dishInd = getInput<int>(getField(dishesIn,k,"dishInd"));
		Real beta = getInput<double>(getField(dishesIn,k,"beta"));
		arr(double) betaSubs = getArrayInput<double>(getField(dishesIn,k,"betaSubs"));
		Real subloglikelihoodOld = getInput<double>(getField(dishesIn,k,"subloglikelihoodOld"));
		Real subloglikelihoodDelta = getInput<double>(getField(dishesIn,k,"subloglikelihoodDelta"));
		bool splittable = getInput<bool>(getField(dishesIn,k,"splittable"));

		Real loglikelihoodOld = getInput<double>(getField(dishesIn,k,"loglikelihoodOld"));
		Real loglikelihoodDelta = getInput<double>(getField(dishesIn,k,"loglikelihoodDelta"));
		bool mergeable = getInput<bool>(getField(dishesIn,k,"mergeable"));

		dishes.emplace(dishInd, dish(dishInd, beta, betaSubs, hyper,
			subloglikelihoodOld, subloglikelihoodDelta, splittable,
			loglikelihoodOld, loglikelihoodDelta, mergeable));
		emptyBeta -= beta;

		for (int d=0; d<D; d++)
			restaurants[d].emplace(dishInd, restaurantDish(0));
	}


	populate_k2z();




	const mxArray* restaurantsIn = getField(inputs,0,"restaurants");
	// loop through the documents/restaurants

	reduction_array<int> NArr(NThreads, K*2, 0);
	reduction_array<int> numTablesArr(NThreads, K, 0);
#ifdef USEFULL
	reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
	reduction_hash2<int> countsArr(NThreads, K*2);
#endif

	#pragma omp parallel for schedule(dynamic)
	for (int d=0; d<D; d++)
	{
		int proc = omp_get_thread_num();
		arr(int) customersIn = getArrayInput<int>(getField(restaurantsIn,d,"customers"));
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
		
		// accumulate the customers in this restaurant
		// number of words / customers
		unsigned int* words = documents.getDocument(d);
		int W = documents.getDocumentLength(d);

		int* customersd = customers[d].data();
		for (int w=0; w<W; w++)
		{
			int word = words[w];
			// transfer the dish assignment
			int c = customersIn[w];
			int dishInd = c/2;
			customersd[w] = c;

			// accumulate the customers at this dish
			//dishes.at(dishInd).add_data_inc_all(c, words[w]);
			(*restaurantsd)[dishInd].addCustomer(c);

			int newk = z2k[dishInd];
			int bin = 2*newk + c%2;
			NArr.reduce_inc(proc, bin);
   		countsArr.reduce_inc(proc, bin, word);
		}

		// accumulate the dish counts for this restaurant
		arr(int) dishCountsIn = getArrayInput<int>(getField(restaurantsIn,d,"dishCounts"));
		int numDishes = mxGetNumberOfElements(getField(restaurantsIn,d,"dishCounts"))/2;
		for (int k=0; k<numDishes; k++)
		{
			int dishInd = dishCountsIn[k*2];
			int dishCount = dishCountsIn[k*2+1];
			(*restaurantsd)[dishInd].setNumTables(dishCount);
			numTablesArr.reduce_add(proc, k, dishCount);
			//dishes[dishInd].add_numTables(dishCount);
		}
	}


   arr(int) fNArr = NArr.final_reduce_add();
   arr(int) fnumTablesArr = numTablesArr.final_reduce_add();
#ifdef USEFULL
   arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif

   #pragma omp parallel for
   for (int k=0; k<K; k++)
   {
#ifdef USEFULL
   	k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   	k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif
   	k2zDishPtrs[k]->add_numTables(fnumTablesArr[k]);
   }










	
}

// --------------------------------------------------------------------------
// -- write_output
// --   creates and writes the output cluster structure
// --------------------------------------------------------------------------
void hdp_subclusters::write_output(mxArray* &plhs)
{
	const char* names[4] = {"gamma","alpha","dishes","restaurants"};
	plhs = mxCreateStructMatrix(1,1,4,names);

	mxWriteField(plhs,0, "gamma", mxCreateScalar(gamma));
	mxWriteField(plhs,0, "alpha", mxCreateScalar(alpha));


	const char* dishNames[9] = {"dishInd","beta","betaSubs","subloglikelihoodOld","subloglikelihoodDelta","splittable",
		"loglikelihoodOld","loglikelihoodDelta","mergeable"};
	mxArray* dishesOut = mxCreateStructMatrix(K,1,9,dishNames);
	mxWriteField(plhs, 0, "dishes", dishesOut);

	for (int k=0; k<K; k++)
	{
		vector<double> betaSubs(2);
		betaSubs[0] = k2zDishPtrs[k]->getBetaL();
		betaSubs[1] = k2zDishPtrs[k]->getBetaR();

		mxWriteField(dishesOut,k, "dishInd", mxCreateScalarINT32(k2zDishPtrs[k]->dishInd));
		mxWriteField(dishesOut,k, "beta", mxCreateScalar(k2zDishPtrs[k]->getBeta()));
		mxWriteField(dishesOut,k, "betaSubs", mxCreateArray<double>(1,2,betaSubs.data()));
		mxWriteField(dishesOut,k, "subloglikelihoodOld", mxCreateScalar(k2zDishPtrs[k]->subloglikelihoodOld));
		mxWriteField(dishesOut,k, "subloglikelihoodDelta", mxCreateScalar(k2zDishPtrs[k]->subloglikelihoodDelta));
		mxWriteField(dishesOut,k, "splittable", mxCreateScalar(k2zDishPtrs[k]->splittable));

		mxWriteField(dishesOut,k, "loglikelihoodOld", mxCreateScalar(k2zDishPtrs[k]->loglikelihoodOld));
		mxWriteField(dishesOut,k, "loglikelihoodDelta", mxCreateScalar(k2zDishPtrs[k]->loglikelihoodDelta));
		mxWriteField(dishesOut,k, "mergeable", mxCreateScalar(k2zDishPtrs[k]->mergeable));
	}

	const char* namesRestaurants[2] = {"customers","dishCounts"};
	mxArray* restaurantsOut = mxCreateStructMatrix(D,1,2,namesRestaurants);
	mxWriteField(plhs, 0, "restaurants", restaurantsOut);

	vector<int> dishCounts(K*2);
	for (int d=0; d<D; d++)
	{
		mxWriteField(restaurantsOut, d, "customers", mxCreateArrayINT32<int>(1,customers[d].size(), customers[d].data()));

		int count = 0;
		for (int k=0; k<K; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			int numTables = restaurants[d][dishInd].getNumTables();
			if (numTables>0)
			{
				dishCounts[count*2] = dishInd;
				dishCounts[count*2+1] = numTables;
				count++;
			}
		}
		mxWriteField(restaurantsOut, d, "dishCounts", mxCreateArrayINT32<int>(2,count, dishCounts.data()));
	}
}

void hdp_subclusters::populate_k2z()
{
	K = dishes.size();
	int k = 0;
	for (dishIterator dishIt=dishes.begin(); dishIt!=dishes.end(); dishIt++, k++)
	{
		int dishInd = dishIt->first;
		k2zDishPtrs[k] = &(dishIt->second);
		z2k[dishInd] = k;
	}
}

int hdp_subclusters::getK()
{
	return K;
}

void hdp_subclusters::getnhist(vector<int>& mvals)
{
	mvals.resize(K);
	#pragma omp parallel for
	for (int k=0; k<K; k++)
		mvals[k] = k2zDishPtrs[k]->getN();
}


Real hdp_subclusters::joint_loglikelihood()
{
	// calculate the joint likelihood
	int T = 0;
	Real loglikelihood = 0;

	// p(\beta,z)
	loglikelihood = K*loggamma;
	loglikelihood += (gamma-1)*fastlog(emptyBeta);
	#pragma omp parallel for reduction(+:loglikelihood)
	for (int k=0; k<K; k++)
	{
		Real beta = k2zDishPtrs[k]->getBeta();
		Real N = k2zDishPtrs[k]->getN();
		loglikelihood += -fastlog(beta);
	}

	#pragma omp parallel for reduction(+:loglikelihood)
	for (int d=0; d<D; d++)
	{
		unsigned int W = documents.getDocumentLength(d);
		loglikelihood += gsl_sf_lngamma(alpha) - gsl_sf_lngamma(alpha + W);
		for (int k=0; k<K; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			Real beta = k2zDishPtrs[k]->getBeta();
			int N = restaurants[d][dishInd].getNumCustomers();
			if (N>0 && beta>0) loglikelihood += gsl_sf_lngamma(N + alpha*beta) - gsl_sf_lngamma(alpha*beta);
		}
	}

	// p(x|z)
	#pragma omp parallel for schedule(guided) reduction(+:loglikelihood)
	for (int k=0; k<K; k++)
		loglikelihood += k2zDishPtrs[k]->data_loglikelihood_marginalized();

	return loglikelihood;
}
Real hdp_subclusters::HOW_loglikelihood()
{
	// calculate the HOW loglikelihood
	Real loglikelihood = 0;
	int count = 0;
	#pragma omp parallel for schedule(dynamic) reduction(+:loglikelihood,count)
	for (int d=0; d<D; d++)
	{
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
		unsigned int W = documents.getDocumentLengthHO(d);
		unsigned int* words = documents.getDocumentHO(d);
		for (int w=0; w<W; w++)
		{
			Real tlogp = -mxGetInf();
			Real totalLogstick = -mxGetInf();
			int word = words[w];
			for (int k=0; k<K; k++)
			{
				int dishInd = k2zDishPtrs[k]->dishInd;
				Real logstick = (*restaurantsd)[dishInd].getLogstickMarginalized(alpha, *(k2zDishPtrs[k]));
				totalLogstick = logsumexp(totalLogstick, logstick);
				tlogp = logsumexp(tlogp, logstick + k2zDishPtrs[k]->predictive_loglikelihood_marginalized(word));
			}
			loglikelihood += tlogp - totalLogstick;

			count++;
		}
	}
	loglikelihood /= count;
	return loglikelihood;
}

void hdp_subclusters::deleteDish(int dishInd)
{
	// make sure to manually call populate_k2z() after all deletions
	// have to delete from dishes and restaurants
	emptyBeta += dishes[dishInd].getBeta();
	dishes.erase(dishInd);

	for (int d=0; d<D; d++)
		restaurants[d].erase(dishInd);

	#ifdef DEBUG
	for (int d=0; d<D; d++)
	{
		unsigned int* words = documents.getDocument(d);
      unsigned int W = documents.getDocumentLength(d);
      int* customersd = customers[d].data();

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      	if (customersd[w]/2 == dishInd)
      		mexErrMsgTxt("trying to delete a dish with customers at it\n");
	}
	#endif
}

void hdp_subclusters::sample_thetas()
{
	// the actual sampling is done in parallel here
	for (dishIterator dishIt=dishes.begin(); dishIt!=dishes.end(); dishIt++)
		dishIt->second.sample();
}
void hdp_subclusters::sample_pis()
{
	// sample the pi's
	#pragma omp parallel for
	for (int d=0; d<D; d++)
	{
		int proc = omp_get_thread_num();
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
		Real total = gsl_ran_gamma(rArr[proc], alpha*emptyBeta, 1);
		for (rdishIterator rdishIt=restaurantsd->begin(); rdishIt!=restaurantsd->end(); rdishIt++)
			total += rdishIt->second.sampleLogstick(rArr[proc], alpha, dishes[rdishIt->first]);
		total = fastlog(total);
		for (rdishIterator rdishIt=restaurantsd->begin(); rdishIt!=restaurantsd->end(); rdishIt++)
			rdishIt->second.normalizeLogstick(total);
	}
}
void hdp_subclusters::sample_params()
{
	// sample pi_j|beta and theta
	sample_thetas();
	sample_pis();
}
	


// OPTION2 : split based on p(z)
void hdp_subclusters::sample_dishes2()
{
	// clear out the counts and data
	#pragma omp parallel for
	for (int d=0; d<D; d++)
	{
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
		for (rdishIterator rdishIt=restaurantsd->begin(); rdishIt!=restaurantsd->end(); rdishIt++)
			rdishIt->second.clearNumCustomers();
	}

	#pragma omp parallel for
	for (int k=0; k<K; k++)
		k2zDishPtrs[k]->empty_params();

	// temporary reduction arrays
	reduction_array<int> NArr(NThreads, K*2, 0);
#ifdef USEFULL
	reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
	reduction_hash2<int> countsArr(NThreads, K*2);
#endif

	reduction_array<Real> likelihoodArr(NThreads, K, 0);
	reduction_array<Real> likelihoodzArr(NThreads, K, 0);
	reduction_array<Real> likelihoodzMtx(NThreads, K*K, 0);


	for (int proc=0; proc<NThreads; proc++)
	{
		arr(int) tprobsArrIndices = probsArrIndices[proc];
		for (int k=0; k<K; k++)
		{
			tprobsArrIndices[k] = k2zDishPtrs[k]->dishInd;
		}
	}
	arr(categorical*) paramPtrs = allocate_memory<categorical*>(K);
	for (int k=0; k<K; k++)
		paramPtrs[k] = &(k2zDishPtrs[k]->main.param.param);


	#pragma omp parallel for schedule(dynamic)
   for (int d=0; d<D; d++)
   {
   	int proc = omp_get_thread_num();

   	arr(Real) tprobsArr = probsArr[proc];
   	arr(int) tprobsArrIndices = probsArrIndices[proc];

      unsigned int* words = documents.getDocument(d);
      unsigned int W = documents.getDocumentLength(d);
      int* customersd = customers[d].data();
      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

      vector<Real> logsticks(K);
      for (int k=0; k<K; k++)
      	logsticks[k] = (*restaurantsd)[k2zDishPtrs[k]->dishInd].getLogstick();

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      {
      	unsigned int word = words[w];

      	// go through and calculate the loglikelihoods for each available topic
      	Real maxProb = -mxGetInf();
      	for (int k=0; k<K; k++)
      	{
   			//int dishInd = k2zDishPtrs[k]->dishInd;
   			//tprobsArrIndices[k] = dishInd;
   			//tprobsArr[k] = logsticks[k] + k2zDishPtrs[k]->predictive_loglikelihood(word);
   			//maxProb = max(maxProb, tprobsArr[k]);
   			tprobsArr[k] = logsticks[k] + paramPtrs[k]->predictive_loglikelihood(word);
   			maxProb = max(maxProb, tprobsArr[k]);
      	}

      	Real totalProb = convert_logcategorical(tprobsArr, K, maxProb);
      	int newk = sample_categorical(tprobsArr, K, totalProb, rArr[proc]);
      	//int newk = customersd[w]/2;
			int dishInd = tprobsArrIndices[newk];

			// left or right?
			Real logpsub, logpzsub;
			int subInd = (*restaurantsd)[dishInd].sampleSubDish(rArr[proc], dishes[dishInd], word, logpsub, logpzsub);
			int c = 2*dishInd + subInd;
			int bin = 2*newk + subInd;

			// accumulate stats
   		customersd[w] = c;
   		(*restaurantsd)[dishInd].addCustomer(c);

   		NArr.reduce_inc(proc, bin);
   		countsArr.reduce_inc(proc, bin, word);
			likelihoodArr.reduce_add(proc, newk, logpsub);
			likelihoodzArr.reduce_add(proc, newk, logpzsub);

			// tprobsArr = log p(x|t)
			// tprobsArr = log p(x|t) - maxProb
			// trpobsArr = exp(log(p(x|t) - maxProb) = p(x|t)*exp(-maxProb)
			// log(p(x|t1)+p(x|t2)) = log( [tprobsArr(1)/exp(-maxProb)] + [tprobsArr(2)/exp(-maxProb)] )
			// log(p(x|t1)+p(x|t2)) = log( [tprobsArr(1) + tprobsArr(2)]/exp(-maxProb) )
			// log(p(x|t1)+p(x|t2)) = log(tprobsArr(1)+tprobsArr(2)) + maxProb
			Real pz = tprobsArr[newk];
			for (int k=0; k<K; k++)
			{
				Real temp = pz;
				if (k!=newk)
					temp += tprobsArr[k];
				temp = fastlog(temp) + maxProb;
				likelihoodzMtx.reduce_add(proc, newk + k*K, temp);
			}
	   }
   }
   deallocate_memory(paramPtrs);

   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
   arr(Real) flikelihoodArr = likelihoodArr.final_reduce_add();
   arr(Real) flikelihoodzArr = likelihoodzArr.final_reduce_add();
   arr(Real) flikelihoodzMtx = likelihoodzMtx.final_reduce_add();
   bool deleted = false;

   logpzMtx.resize(K*K);
   memcpy(logpzMtx.data(), flikelihoodzMtx, sizeof(Real)*K*K);

   #pragma omp parallel for reduction(||:deleted)
   for (int k=0; k<K; k++)
   {
   	bool tempDeleted = true;
   	int numCustomers = 0;
   	for (int s=0; s<2; s++)
   	{
   		tempDeleted = tempDeleted && fNArr[2*k+s]==0;
   		numCustomers += fNArr[2*k+s];
   	}
   	deleted = deleted || tempDeleted;
#ifdef USEFULL
   	k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   	k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif

   	Real newsubloglikelihoodDelta = (flikelihoodArr[k]/numCustomers) - k2zDishPtrs[k]->subloglikelihoodOld;
      if (newsubloglikelihoodDelta<0)// && likelihoodDelta[m]<0)
         k2zDishPtrs[k]->splittable = true;
      k2zDishPtrs[k]->subloglikelihoodDelta = newsubloglikelihoodDelta;
      k2zDishPtrs[k]->subloglikelihoodOld = (flikelihoodArr[k]/numCustomers);

      Real newloglikelihoodDelta = (flikelihoodzMtx[k+k*K]/numCustomers) - k2zDishPtrs[k]->loglikelihoodOld;
      if (newloglikelihoodDelta<0)// && likelihoodDelta[m]<0)
         k2zDishPtrs[k]->mergeable = true;
      k2zDishPtrs[k]->loglikelihoodDelta = newloglikelihoodDelta;
      k2zDishPtrs[k]->loglikelihoodOld = (flikelihoodzMtx[k+k*K]/numCustomers);

      k2zDishPtrs[k]->logpz = flikelihoodzArr[k];
   }

   if (deleted)
   {
   	int oldK = K;
   	vector<int> oldz2k(z2k);
   	for (int k=0; k<K; k++)
   		if (k2zDishPtrs[k]->getN()==0)
   			deleteDish(k2zDishPtrs[k]->dishInd);
		populate_k2z();

		vector<Real> oldlogpzMtx = logpzMtx;
		logpzMtx.resize(K*K);
		for (int k1=0; k1<K; k1++) for (int k2=0; k2<K; k2++)
		{
			int dishInd1 = k2zDishPtrs[k1]->dishInd;
			int dishInd2 = k2zDishPtrs[k2]->dishInd;
			int oldk1 = oldz2k[dishInd1];
			int oldk2 = oldz2k[dishInd2];
			logpzMtx[k1*K + k2] = oldlogpzMtx[oldk1*oldK + oldk2];
		}

		for (int k=0; k<K; k++)
			k2zDishPtrs[k]->set_unsplittable();
   }
}


// OPTION 3
void hdp_subclusters::sample_dishes3()
{
// clear out the counts and data
	#pragma omp parallel for
	for (int d=0; d<D; d++)
	{
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
		for (rdishIterator rdishIt=restaurantsd->begin(); rdishIt!=restaurantsd->end(); rdishIt++)
			rdishIt->second.clearNumCustomers();
	}

	#pragma omp parallel for
	for (int k=0; k<K; k++)
		k2zDishPtrs[k]->empty_params();

	// temporary reduction arrays
	reduction_array<int> NArr(NThreads, K*2, 0);
#ifdef USEFULL
	reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
	reduction_hash2<int> countsArr(NThreads, K*2);
#endif

	reduction_array<Real> likelihoodArr(NThreads, K, 0);
	#pragma omp parallel for schedule(dynamic)
   for (int d=0; d<D; d++)
   {
   	int proc = omp_get_thread_num();

   	arr(Real) tprobsArr = probsArr[proc];
   	arr(int) tprobsArrIndices = probsArrIndices[proc];

      unsigned int* words = documents.getDocument(d);
      unsigned int W = documents.getDocumentLength(d);
      int* customersd = customers[d].data();
      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

      vector<Real> logsticks(K);
      for (int k=0; k<K; k++)
      	logsticks[k] = (*restaurantsd)[k2zDishPtrs[k]->dishInd].getLogstick();

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      {
      	unsigned int word = words[w];

      	// go through and calculate the loglikelihoods for each available topic
      	Real maxProb = -mxGetInf();
      	for (int k=0; k<K; k++)
      	{
   			int dishInd = k2zDishPtrs[k]->dishInd;
   			tprobsArrIndices[k] = dishInd;
   			tprobsArr[k] = logsticks[k] + k2zDishPtrs[k]->predictive_loglikelihood(word);
   			maxProb = max(maxProb, tprobsArr[k]);
      	}

      	Real totalProb = convert_logcategorical(tprobsArr, K, maxProb);
      	int newk = sample_categorical(tprobsArr, K, totalProb, rArr[proc]);
			int dishInd = tprobsArrIndices[newk];

			// left or right?
			Real logpsub, logpzsub;
			int subInd = (*restaurantsd)[dishInd].sampleSubDish(rArr[proc], dishes[dishInd], word, logpsub, logpzsub);
			int c = 2*dishInd + subInd;
			int bin = 2*newk + subInd;

			// accumulate stats
   		customersd[w] = c;
   		(*restaurantsd)[dishInd].addCustomer(c);

   		NArr.reduce_inc(proc, bin);
   		countsArr.reduce_inc(proc, bin, word);
			likelihoodArr.reduce_add(proc, newk, logpsub);
	   }
   }

   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
   arr(Real) flikelihoodArr = likelihoodArr.final_reduce_add();
   bool deleted = false;

   #pragma omp parallel for reduction(||:deleted)
   for (int k=0; k<K; k++)
   {
   	bool tempDeleted = true;
   	int numCustomers = 0;
   	for (int s=0; s<2; s++)
   	{
   		tempDeleted = tempDeleted && fNArr[2*k+s]==0;
   		numCustomers += fNArr[2*k+s];
   	}
   	deleted = deleted || tempDeleted;
#ifdef USEFULL
   	k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   	k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif

   	Real newsubloglikelihoodDelta = (flikelihoodArr[k]/numCustomers) - k2zDishPtrs[k]->subloglikelihoodOld;
      if (newsubloglikelihoodDelta<0)// && likelihoodDelta[m]<0)
         k2zDishPtrs[k]->splittable = true;
      k2zDishPtrs[k]->subloglikelihoodDelta = newsubloglikelihoodDelta;
      k2zDishPtrs[k]->subloglikelihoodOld = (flikelihoodArr[k]/numCustomers);
   }

   if (deleted)
   {
   	int oldK = K;
   	vector<int> oldz2k = z2k;
   	for (int k=0; k<K; k++)
   		if (k2zDishPtrs[k]->getN()==0)
   			deleteDish(k2zDishPtrs[k]->dishInd);
		populate_k2z();

		for (int k=0; k<K; k++)
			k2zDishPtrs[k]->set_unsplittable();
   }
}




// sample the number of tables for each dish in each restaurant
void hdp_subclusters::sample_counts()
{
	reduction_array<int> NTablesArr(NThreads, 3*K, 0);

	#pragma omp parallel for
	for (int d=0; d<D; d++)
	{
		int proc = omp_get_thread_num();

		for (int k=0; k<K; k++)
		{
			dish* dishPtr = k2zDishPtrs[k];
			int dishInd = dishPtr->dishInd;
			rdishIterator rdishIt = restaurants[d].find(dishInd);

			Real ab = alpha * dishPtr->getBeta();
			int numTables = rdishIt->second.sampleNumTables(rArr[proc], ab, logstirling, logstirlingW);
			NTablesArr.reduce_add(proc, 3*k, numTables);

			Real abl = alpha * dishPtr->getBetaL();
			int numTablesL = rdishIt->second.sampleNumTablesL(rArr[proc], abl, logstirling, logstirlingW);
			NTablesArr.reduce_add(proc, 3*k+1, numTablesL);

			Real abr = alpha * dishPtr->getBetaR();
			int numTablesR = rdishIt->second.sampleNumTablesR(rArr[proc], abr, logstirling, logstirlingW);
			NTablesArr.reduce_add(proc, 3*k+2, numTablesR);
		}
	}

	for (int k=0; k<K; k++)
		if (k2zDishPtrs[k]->getN()==0)
			mexErrMsgTxt("Sample Counts Error\n");

	arr(int) fNTablesArr = NTablesArr.final_reduce_add();
   #pragma omp parallel for
   for (int k=0; k<K; k++)
   	k2zDishPtrs[k]->set_numTables(fNTablesArr[3*k], fNTablesArr + 3*k+1);
}


void hdp_subclusters::sample_betas()
{
	Real total = gsl_ran_gamma(rArr[0], gamma, 1); // empty cluster
	#pragma omp parallel for reduction(+:total)
	for (int k=0; k<K; k++)
	{
		total += k2zDishPtrs[k]->sampleBeta(rArr[omp_get_thread_num()], gamma);
	}

	Real totalBeta = 0;
	#pragma omp parallel for reduction(+:totalBeta)
	for (int k=0; k<K; k++)
		totalBeta += k2zDishPtrs[k]->normalizeBeta(total);
	emptyBeta = 1-totalBeta;
}


void hdp_subclusters::sample_hyperparameters()
{
	//mexPrintf("here\n");drawnow();
	// sample alpha first
	for (int it=0; it<1; it++)
	{
		Real sumlogw = 0;
		int sums = 0;
		#pragma omp parallel for reduction(+:sumlogw,sums)
		for (int d=0; d<D; d++)
		{
			int proc = omp_get_thread_num();
			Real W = documents.getDocumentLength(d);
			Real w = gsl_ran_beta(rArr[proc], alpha+1, W);
			Real probs = 1 + (W/alpha);
			int s = (my_rand(rArr[proc]) > 1.0/probs);
			sums += s;
			sumlogw += log(w);
		}
	//mexPrintf("1\n");drawnow();
		int m = 0;
		for (int k=0; k<K; k++)
			m += k2zDishPtrs[k]->getNumTables();

		alpha = gsl_ran_gamma(rArr[0], 1 + m - sums, 1.0/(1.0 - sumlogw));
		//mexPrintf("m=%d, sums=%d, sumlogw=%e\n", m, sums, sumlogw);
		logalpha = log(alpha);


		Real a = 1;
		Real b = 0.1;
		Real eta = gsl_ran_beta(rArr[0], alpha+1, m);
		Real pietaRatio = (a+K-1)/ (m*(b-log(eta)));
		Real pieta = pietaRatio / (1+pietaRatio);
		int s = my_rand(rArr[0]) > pieta;
		gamma = gsl_ran_gamma(rArr[0], a+K+s, 1.0/(b-log(eta)));
		loggamma = log(gamma);
		//mexPrintf("2\n");drawnow();
		//mexPrintf("%e %e\n", alpha, gamma);drawnow();
	}



	/*lambdas.resize(hyper.getD());
	int PL = 100;
	vector<double> possible_lambdas(PL);
	for (pl=0; pl<PL; pl++)
		possible_lambdas[pl] = gsl_ran_gamma(rArr[0], 1, 1);
	*/
}






void hdp_subclusters::propose_splits1()
{
	int Ks = 0;
	for (int k=0; k<K; k++)
		if (always_splittable || k2zDishPtrs[k]->splittable)
			++Ks;

	if (Ks==0)
		return;

   // sample new betas for the split.
   // first find the optimal m count
   vector<int> mnewL(K);
   vector<int> mnewR(K);
   split2_find_bestNumTables(mnewL, mnewR);

   vector<Real> newBetaL(K);
   vector<Real> newBetaR(K);
   split2_sample_betas_bestNumTables(mnewL, mnewR, newBetaL, newBetaR);

   vector<Real> splitCustomerLoglikelihoods(K);
   split2_splitCustomersLoglikelihood(newBetaL, newBetaR, splitCustomerLoglikelihoods);

	vector<bool> do_split(K, false);
   int num_splits = 0;
   Real tempEmptyBeta = emptyBeta;
   //#pragma omp parallel for reduction(+:num_splits)
   for (int k=0; k<K; k++)
   {
      // check to see if the subclusters have converged
      if (always_splittable || k2zDishPtrs[k]->splittable)//paramsChanges[k2z[k]]/((double)params[k2z[k]]->getN())<=0.1)
      {
         dish* dishPtr = k2zDishPtrs[k];
         int dishInd = dishPtr->dishInd;

         int numCustomers = dishPtr->getN();
         int numCustomersL = dishPtr->getNL();
         int numCustomersR = dishPtr->getNR();

         if (numCustomersL==0 || numCustomersR==0)
         	continue;

         Real beta = dishPtr->getBeta();
         Real betaL = newBetaL[k] * beta;
         Real betaR = newBetaR[k] * beta;

   		//mexPrintf("k=%d\t",dishInd);drawnow();
   		Real HR = loggamma;

   		Real temp = 2*(Ks/2);
   		HR += log(temp / (Ks*(Ks+1)/2));

   		HR += gammalnint(mnewL[k]) + gammalnint(mnewR[k]) - gammalnint(mnewL[k]+mnewR[k]);
   		HR += -mnewL[k]*log(betaL) - mnewR[k]*log(betaR) + (mnewL[k]+mnewR[k])*log(beta);
   		//mexPrintf("%e\t", HR);

			HR += splitCustomerLoglikelihoods[k];
   		//mexPrintf("%e\t", HR);

   		//HR -= k2zDishPtrs[k]->logpz;
   		//mexPrintf("%e\t", HR);

   		HR += dishPtr->data_loglikelihood_marginalizedL() + dishPtr->data_loglikelihood_marginalizedR();
   		HR -= dishPtr->data_loglikelihood_marginalized();
   		//mexPrintf("%e\n", HR);

         if ((HR>0 || my_rand(rArr[0]) < fastexp(HR)))
         {
         	//mexPrintf("split %d\n", dishInd);
            do_split[k] = true;
            num_splits++;
         }
      }
   }

   if (num_splits>0)
   {
      //mexPrintf("NumSplits=%d\n", num_splits);
      int temp_num_splits = num_splits;
      // figure out which labels to split into and prepare simple split stuff
      vector<int> split_into(K);
      vector<int> split_into_k(K);

      int k = 0;
      for (k; k<K; k++)
         if (do_split[k])
            break;
      vector<Real> tempBetaSubs(2, 0.5);
      for (int nzh=0; nzh<D; nzh++)
      {
         if (dishes.count(nzh)==0)
         {
         	int dishInd1 = k2zDishPtrs[k]->dishInd;
            int dishInd2 = nzh;
            split_into[k] = nzh;
            split_into_k[k] = K+(num_splits-temp_num_splits);

            Real oldBeta = k2zDishPtrs[k]->getBeta();
            Real betaL = newBetaL[k] * oldBeta;
            Real betaR = newBetaR[k] * oldBeta;
            dishes[dishInd2] = dish(dishInd2, betaR, tempBetaSubs.data(), hyper,-mxGetInf(),mxGetInf(),false,-mxGetInf(),mxGetInf(),false);
				dishes[dishInd1].setBeta(betaL);

            // add the dish to all restaurants
            for (int d=0; d<D; d++)
            {
            	restaurants[d][dishInd1].clearNumCustomers();
            	restaurants[d][dishInd2] = restaurantDish(0);
            	restaurants[d][dishInd2].split_fromSub(1, restaurants[d][dishInd1]);
            	restaurants[d][dishInd1].split_fromSub(0, restaurants[d][dishInd1]);
            }

            // transfer over the theta's and beta's
            dishes[dishInd2].split_fromSub(1, *(k2zDishPtrs[k]));

         	// mark them as unsplittable
            k2zDishPtrs[k]->subloglikelihoodOld = -mxGetInf();
            k2zDishPtrs[k]->subloglikelihoodDelta = mxGetInf();
            k2zDishPtrs[k]->splittable = false;

            // transfer over the theta's and beta's
	         dishes[dishInd1].split_fromSub(0, *(k2zDishPtrs[k]));

         	temp_num_splits--;
         	// do the next ones if there are any
            if (temp_num_splits>0)
            {
               for (k=k+1; k<K; k++)
                  if (do_split[k])
                     break;
            }
            else
               break;
         }
      }

      // correct the labels
      // temporary reduction array variables
      reduction_array<int> NArr(NThreads, (K+num_splits*(2-1))*2, 0);
#ifdef USEFULL
		reduction_array2<int> countsArr(NThreads, (K+num_splits)*2, hyper.getD(), 0);
#else
		reduction_hash2<int> countsArr(NThreads, (K+num_splits)*2);
#endif

		#pragma omp parallel for schedule(dynamic)
	   for (int d=0; d<D; d++)
	   {
	   	int proc = omp_get_thread_num();

	      unsigned int* words = documents.getDocument(d);
	      unsigned int W = documents.getDocumentLength(d);
	      int* customersd = customers[d].data();
	      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

	      for (int w=0; w<W; w++) if (do_split[z2k[customersd[w]/2]])
	      {
	      	int k = z2k[customersd[w]/2];
	      	unsigned int word = words[w];

	      	int s = customersd[w]%2;
				int newk, dishInd;
	      	if (s==0)
	      	{
	      		dishInd = customersd[w]/2;
	      		newk = k;
	      	}
	      	else
	      	{
	      		dishInd = split_into[k];
	      		newk = split_into_k[k];
	      	}
				int isRight = my_rand(rArr[proc])<0.5;
				int newc = 2*dishInd + isRight;
	      	int bin = 2*newk + isRight;

	      	customersd[w] = newc;
	      	(*restaurantsd)[dishInd].addCustomer(newc);
	      	
	      	NArr.reduce_inc(proc, bin);
	      	countsArr.reduce_inc(proc, bin, word);
		   }
	   }

	   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   	arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   	unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
	   //#pragma omp parallel for
	   for (int k=0; k<K; k++) if (do_split[k])
	   {
	   	int dishInd1 = k2zDishPtrs[k]->dishInd;
#ifdef USEFULL
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif
	   	k2zDishPtrs[k]->sample_subdishes();

	   	int dishInd2 = split_into[k];
	   	int newk = split_into_k[k];
#ifdef USEFULL
	   	dishes[dishInd2].set_stats(fNArr + (2*newk), fcountsArr + (2*newk)*hyper.getD());
#else
   		dishes[dishInd2].set_stats(fNArr + (2*newk), fcountsArr + (2*newk));
#endif
   		dishes[dishInd2].sample_subdishes();
	   	// need to sample 
	   	//	1) a new set of theta's for the subdishes
	   	// 2) a new set of pi's for all restaurants
	   }
	   populate_k2z();
	   sample_pis();
	   // sample counts is done next step anyway
   }
}










// ===================================================================================================
// ===================================================================================================
// ===================================================================================================



// find the best number of tables for splitmerge2 moves
void hdp_subclusters::merge2_find_bestNumTables(vector<int>& merge_proposals, int Ks,
	vector<int>& moldL, vector<int>& moldR)
{
	int oldBetaAvg = alpha*(1.0/K)*100 - 0.5;
   if (oldBetaAvg>=100)
   	mexErrMsgTxt("bestNumTables not big enough\n");
   oldBetaAvg *= bestNumTablesN;

   // loop through each document and find the best table count
   for (int it=0; it<Ks/2; it++)
   {
   	int k1 = merge_proposals[it*2];
   	int k2 = merge_proposals[it*2+1];
   	int dishInd1 = k2zDishPtrs[k1]->dishInd;
   	int dishInd2 = k2zDishPtrs[k2]->dishInd;
   	int moldlk = 0;
   	int moldrk = 0;
   	#pragma omp parallel for reduction(+:moldlk,moldrk)
	   for (int d=0; d<D; d++)
	   {
	   	int numCustomers1 = restaurants[d][dishInd1].getNumCustomers();
	   	int numCustomers2 = restaurants[d][dishInd2].getNumCustomers();
			moldlk += bestNumTables[numCustomers1 + oldBetaAvg];
			moldrk += bestNumTables[numCustomers2 + oldBetaAvg];
	   }
	   moldL[it] = moldlk;
	   moldR[it] = moldrk;
   }
}

void hdp_subclusters::merge2_mergeCustomersLoglikelihood(vector<int>& merge_proposals,
	int Ks, vector<Real>& loglikelihoods)
{
	for (int it=0; it<Ks/2; it++)
	{
		int k1 = merge_proposals[2*it];
		int k2 = merge_proposals[2*it+1];
		Real betaL = k2zDishPtrs[k1]->getBeta();
		Real betaR = k2zDishPtrs[k2]->getBeta();
		Real beta = betaL + betaR;
		Real ab = alpha*beta;
		Real abl = alpha*betaL;
		Real abr = alpha*betaR;

		int dishInd1 = k2zDishPtrs[k1]->dishInd;
		int dishInd2 = k2zDishPtrs[k2]->dishInd;
		Real loglikelihood = 0;
		#pragma omp parallel for reduction(+:loglikelihood)
		for (int d=0; d<D; d++)
		{
			int proc = omp_get_thread_num();
			int numCustomersL = restaurants[d][dishInd1].getNumCustomers();
			int numCustomersR = restaurants[d][dishInd2].getNumCustomers();
			int numCustomers = numCustomersL + numCustomersR;
			if (numCustomers>0) loglikelihood += gsl_sf_lngamma(ab + numCustomers) - gsl_sf_lngamma(ab);
			if (numCustomersL>0) loglikelihood += gsl_sf_lngamma(abl) - gsl_sf_lngamma(abl + numCustomersL);
			if (numCustomersR>0) loglikelihood += gsl_sf_lngamma(abr) - gsl_sf_lngamma(abr + numCustomersR);
		}
		loglikelihoods[it] = loglikelihood;
	}
}
void hdp_subclusters::propose_merges2()
{
	int oKs = 0;
	for (int k=0; k<K; k++)
		if (always_splittable || k2zDishPtrs[k]->splittable)
			oKs++;
	//oKs /= 20;
	//oKs = max(oKs,1);
	int numMerges = 0;
	vector<int> merge_with(K,-1);
	vector<int> merge_proposals(K);
	for (int mergeit=0; mergeit<oKs; mergeit++)
	{
		int Ks = 0;
		for (int k=0; k<K; k++)
			if ((always_splittable || k2zDishPtrs[k]->splittable) && merge_with[k]<0)
				merge_proposals[Ks++] = k;
		
		// do a Knuth shuffle
		for (int ks=0; ks<Ks; ks++)
		{
			int swapk = ks + my_randi(rArr[0]) % (Ks-ks);
			int temp = merge_proposals[ks];
			merge_proposals[ks] = merge_proposals[swapk];
			merge_proposals[swapk] = temp;
		}

		vector<int> moldL(Ks/2);
		vector<int> moldR(Ks/2);
		merge2_find_bestNumTables(merge_proposals, Ks, moldL, moldR);

		vector<Real> mergeCustomerLoglikelihoods(K);
	   merge2_mergeCustomersLoglikelihood(merge_proposals, Ks, mergeCustomerLoglikelihoods);

	   //#pragma omp parallel for reduction(+:numMerges)
	   for (int it=0; it<Ks/2; it++)
	   {
	   	int proc = omp_get_thread_num();
	   	int km = merge_proposals[2*it];
	   	int kn = merge_proposals[2*it+1];
	   	int dishInd1 = k2zDishPtrs[km]->dishInd;
	   	int dishInd2 = k2zDishPtrs[kn]->dishInd;
	   	Real betaL = k2zDishPtrs[km]->getBeta();
	   	Real betaR = k2zDishPtrs[kn]->getBeta();
	   	Real beta = betaL + betaR;

	   	//mexPrintf("1(%d,%d)\t", dishInd1, dishInd2);
	   	Real HR = -loggamma;

	   	Real temp = 2*(Ks/2);
	   	HR -= log(oKs*temp / (Ks*(Ks-1)/2));

	   	HR += -gammalnint(moldL[it]) - gammalnint(moldR[it]) + gammalnint(moldL[it]+moldR[it]);
			HR += moldL[it]*log(betaL) + moldR[it]*log(betaR) - (moldL[it]+moldR[it])*log(beta);
			//mexPrintf("%e\t", HR);

			HR += mergeCustomerLoglikelihoods[it];
			//mexPrintf("%e\t", HR);

			HR += logpzMtx[km+km*K] + logpzMtx[kn+kn*K] - logpzMtx[km+kn*K] - logpzMtx[kn+km*K];
			//mexPrintf("%e\t", HR);

			HR += k2zDishPtrs[km]->data_loglikelihood_marginalized_testmerge(*(k2zDishPtrs[kn]));
			HR += -k2zDishPtrs[km]->data_loglikelihood_marginalized() - k2zDishPtrs[kn]->data_loglikelihood_marginalized();
			//mexPrintf("%e\n", HR);

	      if ((HR>0 || my_rand(rArr[proc]) < fastexp(HR)))
	      {
	         merge_with[km] = km;
	         merge_with[kn] = km;
	         numMerges++;
	      }
	   }
	}


   if (numMerges>0)
   {
   	//mexPrintf("NumMerges=%d\n", numMerges);
   	#pragma omp parallel for
		for (int d=0; d<D; d++)
		{
			unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
			for (int k=0; k<K; k++)
				if (merge_with[k]>=0)
					(*restaurantsd)[k2zDishPtrs[k]->dishInd].clearNumCustomers();
		}

		#pragma omp parallel for
		for (int k=0; k<K; k++) if (merge_with[k]>=0)
			k2zDishPtrs[k]->empty_params();


		for (int kn=0; kn<K; kn++) if (merge_with[kn]>=0 && merge_with[kn]!=kn)
		{
			int dishInd2 = k2zDishPtrs[kn]->dishInd;
      	int km = merge_with[kn];
      	int dishInd1 = k2zDishPtrs[km]->dishInd;
         // merge the dish
         k2zDishPtrs[km]->merge_with(*(k2zDishPtrs[kn]));

         // merge the restaurant dishes
         #pragma omp parallel for
         for (int d=0; d<D; d++)
      		restaurants[d][dishInd1].merge_with(restaurants[d][dishInd2]);
      }


		reduction_array<int> NArr(NThreads, K*2, 0);
#ifdef USEFULL
	reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
	reduction_hash2<int> countsArr(NThreads, K*2);
#endif
		#pragma omp parallel for schedule(dynamic)
	   for (int d=0; d<D; d++)
	   {
	   	int proc = omp_get_thread_num();
	      unsigned int* words = documents.getDocument(d);
	      unsigned int W = documents.getDocumentLength(d);
	      int* customersd = customers[d].data();
	      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

	      // too hard to do random order of words... just do it in order
	      for (int w=0; w<W; w++)
	      {
	      	int k = z2k[customersd[w]/2];
	      	unsigned int word = words[w];

	      	if (merge_with[k]>=0)
				{
					int newk = merge_with[k];
					int dishInd = k2zDishPtrs[newk]->dishInd;

					int subInd = merge_with[k]!=k;
					int c = 2*dishInd + subInd;
					int bin = 2*newk + subInd;

					// accumulate stats
   				customersd[w] = c;
   				(*restaurantsd)[dishInd].addCustomer(c);

	   			NArr.reduce_inc(proc, bin);
   				countsArr.reduce_inc(proc, bin, word);
				}
	   	}
   	}

	   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   	arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   	unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif

	   #pragma omp parallel for
	   for (int k=0; k<K; k++) if (merge_with[k]>=0 && merge_with[k]==k)
	   {
	   	int numCustomers = 0;
   		for (int s=0; s<2; s++)
	   		numCustomers += fNArr[2*k+s];
#ifdef USEFULL
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif

   		k2zDishPtrs[k]->set_unsplittable();
   		k2zDishPtrs[k]->set_unmergeable();
   	}

   	for (int k=0; k<K; k++)
      	if (merge_with[k]>=0 && merge_with[k]!=k)
      		deleteDish(k2zDishPtrs[k]->dishInd);

      populate_k2z();
   }


   /*for (int k=0; k<K; k++)
   	if (k2zDishPtrs[k]->getN()==0)
   		deleteDish(k2zDishPtrs[k]->dishInd);
	populate_k2z();*/
}



// find the best number of tables for splitmerge2 moves
void hdp_subclusters::split2_find_bestNumTables(vector<int>& mnewL,
	vector<int>& mnewR)
{
	int oldBetaAvg = alpha*(1.0/K)*100 - 0.5;
   int newBetaAvg = alpha*(1.0/(K+1))*100 - 0.5;
   if (oldBetaAvg>=100 || newBetaAvg>=100)
   	mexErrMsgTxt("bestNumTables not big enough\n");
   oldBetaAvg *= bestNumTablesN;
   newBetaAvg *= bestNumTablesN;

   // loop through each document and find the best table count
   for (int k=0; k<K; k++) if (always_splittable || k2zDishPtrs[k]->splittable)
   {
   	int dishInd = k2zDishPtrs[k]->dishInd;
   	int mnewkl = 0;
   	int mnewkr = 0;
   	#pragma omp parallel for reduction(+:mnewkl,mnewkr)
	   for (int d=0; d<D; d++)
	   {
			mnewkl += bestNumTables[restaurants[d][dishInd].getNumCustomersL() + newBetaAvg];
			mnewkr += bestNumTables[restaurants[d][dishInd].getNumCustomersR() + newBetaAvg];
	   }
	   mnewL[k] = mnewkl;
	   mnewR[k] = mnewkr;

	   /*if ((mnewL[k]==0 && k2zDishPtrs[k]->getNL()>0) || (mnewR[k]==0 && k2zDishPtrs[k]->getNR()>0))
	   {
	   	mexPrintf("=============================\n");
	   	int t = 0;
	   	int tl = 0;
	   	int tr = 0;
	   	for (int d=0; d<D; d++)
	   	{
	   		int numCustomers = restaurants[d][dishInd].getNumCustomers();
	   		int numCustomersL = restaurants[d][dishInd].getNumCustomersL();
	   		int numCustomersR = restaurants[d][dishInd].getNumCustomersR();

				t += numCustomers;
				tl += numCustomersL;
				tr += numCustomersR;
	   	}

	   	mexPrintf("rest custs %d %d %d\n", t, tl, tr);
	   	mexPrintf("dish custs %d %d %d\n", k2zDishPtrs[k]->getN(), k2zDishPtrs[k]->getNL(), k2zDishPtrs[k]->getNR());
	   	mexPrintf("newBetaAvg = %e\n", newBetaAvg);
	   }*/
   }
}
void hdp_subclusters::split2_sample_betas_bestNumTables(vector<int> &mnewL, vector<int>& mnewR,
	vector<Real> &betaL, vector<Real> &betaR)
{
   #pragma omp parallel for
   for (int k=0; k<K; k++) if (always_splittable || k2zDishPtrs[k]->splittable)
   {
   	int proc = omp_get_thread_num();
   	betaL[k] = gsl_ran_gamma(rArr[proc], mnewL[k], 1);
   	betaR[k] = gsl_ran_gamma(rArr[proc], mnewR[k], 1);
   	Real total = betaL[k] + betaR[k];
   	betaL[k] /= total;
   	betaR[k] /= total;

   	/*if (betaL[k]!=betaL[k] || betaR[k]!=betaR[k])
   	{
   		mexPrintf("Tables: %d %d %d\n", mnewL[k], mnewR[k], k2zDishPtrs[k]->getNumTables());
   		mexPrintf("Customers: %d %d %d\n", k2zDishPtrs[k]->getNL(), k2zDishPtrs[k]->getNR(), k2zDishPtrs[k]->getN());
   		mexPrintf("total=%e\n", total); 
   		mexErrMsgTxt("a\n");
   	}*/
   }
}

void hdp_subclusters::split2_splitCustomersLoglikelihood(const vector<Real> &newBetaL, const vector<Real> &newBetaR,
   vector<Real>& loglikelihoods)
{
	for (int k=0; k<K; k++) if (always_splittable || k2zDishPtrs[k]->splittable)
	{
		Real beta = k2zDishPtrs[k]->getBeta();
		Real betaL = newBetaL[k]*beta;
		Real betaR = newBetaR[k]*beta;
		Real ab = alpha*beta;
		Real abl = alpha*betaL;
		Real abr = alpha*betaR;

		int dishInd = k2zDishPtrs[k]->dishInd;
		Real loglikelihood = 0;
		#pragma omp parallel for reduction(+:loglikelihood)
		for (int d=0; d<D; d++)
		{
			int proc = omp_get_thread_num();
			restaurantDish* rdishPtr = &(restaurants[d][dishInd]);
			int numCustomers = rdishPtr->getNumCustomers();
			int numCustomersL = rdishPtr->getNumCustomersL();
			int numCustomersR = rdishPtr->getNumCustomersR();

			/*if ((numCustomers>0 && ab<=0) || (numCustomersL>0 && abl<=0) || (numCustomersR>0 && abr<=0) || ab!=ab || abl!=abl || abr!=abr)
			{
				mexPrintf("%d %d %d\n", numCustomers, numCustomersL, numCustomersR);drawnow();
				mexPrintf("%e %e %e\n", ab, abl, abr);drawnow();
				mexErrMsgTxt("a\n");
			}*/

			if (numCustomers>0) loglikelihood += gsl_sf_lngamma(ab) - gsl_sf_lngamma(ab + numCustomers);
			if (numCustomersL>0) loglikelihood += gsl_sf_lngamma(abl + numCustomersL) - gsl_sf_lngamma(abl);
			if (numCustomersR>0) loglikelihood += gsl_sf_lngamma(abr + numCustomersR) - gsl_sf_lngamma(abr);
		}
		loglikelihoods[k] = loglikelihood;
	}
}
void hdp_subclusters::propose_splits2()
{
	int Ks = 0;
	for (int k=0; k<K; k++)
		if (always_splittable || k2zDishPtrs[k]->splittable)
			++Ks;

	if (Ks==0)
		return;

   // sample new betas for the split.
   // first find the optimal m count
   vector<int> mnewL(K);
   vector<int> mnewR(K);
   split2_find_bestNumTables(mnewL, mnewR);

   vector<Real> newBetaL(K);
   vector<Real> newBetaR(K);
   split2_sample_betas_bestNumTables(mnewL, mnewR, newBetaL, newBetaR);

   vector<Real> splitCustomerLoglikelihoods(K);
   split2_splitCustomersLoglikelihood(newBetaL, newBetaR, splitCustomerLoglikelihoods);

	vector<bool> do_split(K, false);
   int num_splits = 0;
   Real tempEmptyBeta = emptyBeta;
   //#pragma omp parallel for reduction(+:num_splits)
   for (int k=0; k<K; k++)
   {
      // check to see if the subclusters have converged
      if (always_splittable || k2zDishPtrs[k]->splittable)//paramsChanges[k2z[k]]/((double)params[k2z[k]]->getN())<=0.1)
      {
         dish* dishPtr = k2zDishPtrs[k];
         int dishInd = dishPtr->dishInd;

         int numCustomers = dishPtr->getN();
         int numCustomersL = dishPtr->getNL();
         int numCustomersR = dishPtr->getNR();

         if (numCustomersL==0 || numCustomersR==0)
         	continue;

         Real beta = dishPtr->getBeta();
         Real betaL = newBetaL[k] * beta;
         Real betaR = newBetaR[k] * beta;

   		//mexPrintf("k=%d\t",dishInd);drawnow();
   		Real HR = loggamma;

   		Real temp = 2*(Ks/2) * (Ks/10);
   		//HR += log(Ks*temp / (Ks*(Ks+1)/2));
   		HR += fastlog(temp / (Ks*(Ks+1)/2));

   		HR += gammalnint(mnewL[k]) + gammalnint(mnewR[k]) - gammalnint(mnewL[k]+mnewR[k]);
   		HR += -mnewL[k]*fastlog(betaL) - mnewR[k]*fastlog(betaR) + (mnewL[k]+mnewR[k])*fastlog(beta);
   		//mexPrintf("%e\t", HR);

			HR += splitCustomerLoglikelihoods[k];
   		//mexPrintf("%e\t", HR);

   		HR -= k2zDishPtrs[k]->logpz;
   		//mexPrintf("%e\t", HR);

   		HR += dishPtr->data_loglikelihood_marginalizedL() + dishPtr->data_loglikelihood_marginalizedR();
   		HR -= dishPtr->data_loglikelihood_marginalized();
   		//mexPrintf("%e\n", HR);

         if ((HR>0 || my_rand(rArr[0]) < fastexp(HR)))
         {
         	//mexPrintf("split %d\n", dishInd);
            do_split[k] = true;
            num_splits++;
         }
      }
   }

   if (num_splits>0)
   {
      //mexPrintf("NumSplits=%d\n", num_splits);
      int temp_num_splits = num_splits;
      // figure out which labels to split into and prepare simple split stuff
      vector<int> split_into(K);
      vector<int> split_into_k(K);

      int k = 0;
      for (k; k<K; k++)
         if (do_split[k])
            break;
      vector<Real> tempBetaSubs(2, 0.5);
      for (int nzh=0; nzh<D; nzh++)
      {
         if (dishes.count(nzh)==0)
         {
         	int dishInd1 = k2zDishPtrs[k]->dishInd;
            int dishInd2 = nzh;
            split_into[k] = nzh;
            split_into_k[k] = K+(num_splits-temp_num_splits);

            Real oldBeta = k2zDishPtrs[k]->getBeta();
            Real betaL = newBetaL[k] * oldBeta;
            Real betaR = newBetaR[k] * oldBeta;
            dishes[dishInd2] = dish(dishInd2, betaR, tempBetaSubs.data(), hyper,-mxGetInf(),mxGetInf(),false,-mxGetInf(),mxGetInf(),false);
            //dishes.insert(std::make_pair<int,dish>(dishInd2, dish(dishInd2, betaR, tempBetaSubs.data(), hyper,-mxGetInf(),mxGetInf(),false,-mxGetInf(),mxGetInf(),false)));
				dishes[dishInd1].setBeta(betaL);

            // add the dish to all restaurants
            for (int d=0; d<D; d++)
            {
            	restaurants[d][dishInd1].clearNumCustomers();
            	restaurants[d][dishInd2] = restaurantDish(0);
            	//restaurants[d].insert(std::make_pair<int,restaurantDish>(dishInd2, restaurantDish(0)));
            	restaurants[d][dishInd2].split_fromSub(1, restaurants[d][dishInd1]);
            	restaurants[d][dishInd1].split_fromSub(0, restaurants[d][dishInd1]);
            }

            // transfer over the theta's and beta's
            dishes[dishInd2].split_fromSub(1, *(k2zDishPtrs[k]));

         	// mark them as unsplittable
         	k2zDishPtrs[k]->set_unsplittable();
         	k2zDishPtrs[k]->set_unmergeable();

            // transfer over the theta's and beta's
	         dishes[dishInd1].split_fromSub(0, *(k2zDishPtrs[k]));

         	temp_num_splits--;
         	// do the next ones if there are any
            if (temp_num_splits>0)
            {
               for (k=k+1; k<K; k++)
                  if (do_split[k])
                     break;
            }
            else
               break;
         }
      }

      // correct the labels
      // temporary reduction array variables
      reduction_array<int> NArr(NThreads, (K+num_splits*(2-1))*2, 0);
#ifdef USEFULL
		reduction_array2<int> countsArr(NThreads, (K+num_splits)*2, hyper.getD(), 0);
#else
		reduction_hash2<int> countsArr(NThreads, (K+num_splits)*2);
#endif

		#pragma omp parallel for schedule(dynamic)
	   for (int d=0; d<D; d++)
	   {
	   	int proc = omp_get_thread_num();

	      unsigned int* words = documents.getDocument(d);
	      unsigned int W = documents.getDocumentLength(d);
	      int* customersd = customers[d].data();
	      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

	      for (int w=0; w<W; w++) if (do_split[z2k[customersd[w]/2]])
	      {
	      	int k = z2k[customersd[w]/2];
	      	unsigned int word = words[w];

	      	int s = customersd[w]%2;
				int newk, dishInd;
	      	if (s==0)
	      	{
	      		dishInd = customersd[w]/2;
	      		newk = k;
	      	}
	      	else
	      	{
	      		dishInd = split_into[k];
	      		newk = split_into_k[k];
	      	}
				int isRight = my_rand(rArr[proc])<0.5;
				int newc = 2*dishInd + isRight;
	      	int bin = 2*newk + isRight;

	      	customersd[w] = newc;
	      	(*restaurantsd)[dishInd].addCustomer(newc);
	      	
	      	NArr.reduce_inc(proc, bin);
	      	countsArr.reduce_inc(proc, bin, word);
		   }
	   }

	   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   	arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   	unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
	   //#pragma omp parallel for
	   for (int k=0; k<K; k++) if (do_split[k])
	   {
	   	int dishInd1 = k2zDishPtrs[k]->dishInd;
#ifdef USEFULL
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif
	   	k2zDishPtrs[k]->sample_subdishes();

	   	int dishInd2 = split_into[k];
	   	int newk = split_into_k[k];
#ifdef USEFULL
	   	dishes[dishInd2].set_stats(fNArr + (2*newk), fcountsArr + (2*newk)*hyper.getD());
#else
   		dishes[dishInd2].set_stats(fNArr + (2*newk), fcountsArr + (2*newk));
#endif
   		dishes[dishInd2].sample_subdishes();
	   	// need to sample 
	   	//	1) a new set of theta's for the subdishes
	   	// 2) a new set of pi's for all restaurants
	   }
	   populate_k2z();
	   sample_pis();
	   // sample counts is done next step anyway
   }
}






// ===================================================================================================
// ===================================================================================================
// ===================================================================================================




// proposes a new z~p(z|~theta,~pi) and return the probability of it happening
Real hdp_subclusters::splitmerge_sample_ztilde(
	vector< unordered_map<int,restaurantDishComps> > &trestaurants,
	vector<dishComps*> &k2ztDishPtrs)
{
	int Knew = k2ztDishPtrs.size();
	// temporary reduction arrays
	reduction_array<int> NArr(NThreads, (Knew), 0);
#ifdef USEFULL
	reduction_array2<int> countsArr(NThreads, Knew, hyper.getD(), 0);
#else
	reduction_hash2<int> countsArr(NThreads, Knew);
#endif
	Real newloglikelihoodz = 0;
	// propose a new z~p(z|~theta,~pi)
	#pragma omp parallel for schedule(dynamic) reduction(+:newloglikelihoodz)
   for (int d=0; d<D; d++)
   {
   	int proc = omp_get_thread_num();

   	arr(Real) tprobsArr = probsArr[proc];
   	arr(int) tprobsArrIndices = probsArrIndices[proc];

      unsigned int* words = documents.getDocument(d);
      unsigned int W = documents.getDocumentLength(d);
      int* tcustomersd = tcustomers[d].data();
      unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      {
      	unsigned int word = words[w];

      	// go through and calculate the loglikelihoods for each available topic
#ifdef USELOGLIKELIHOOD
			Real maxProb = -mxGetInf();
      	for (int k=0; k<Knew; k++)
      	{
   			int dishInd = k2ztDishPtrs[k]->dishInd;
   			tprobsArrIndices[k] = dishInd;
   			tprobsArr[k] = (*trestaurantsd)[dishInd].getLogstick() + k2ztDishPtrs[k]->predictive_loglikelihood(word);
   			maxProb = max(maxProb, tprobsArr[k]);
      	}
      	Real totalProb = convert_logcategorical(tprobsArr, Knew, maxProb);
#else
      	Real maxProb = -mxGetInf();
      	Real totalProb = 0;
      	for (int k=0; k<Knew; k++)
      	{
   			int dishInd = k2ztDishPtrs[k]->dishInd;
   			tprobsArrIndices[k] = dishInd;
   			tprobsArr[k] = (*trestaurantsd)[dishInd].getStick() * k2ztDishPtrs[k]->predictive_likelihood(word);
   			totalProb += tprobsArr[k];
      	}
#endif

      	int newk = sample_categorical(tprobsArr, Knew, totalProb, rArr[proc]);
			int dishInd = tprobsArrIndices[newk];

			int c = 2*dishInd;
			int bin = newk;

			// accumulate stats
   		tcustomersd[w] = c;
   		(*trestaurantsd)[dishInd].addCustomer();

   		NArr.reduce_inc(proc, bin);
   		countsArr.reduce_inc(proc, bin, word);
			newloglikelihoodz += fastlog(tprobsArr[newk] / totalProb);
	   }
   }
   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
   #pragma omp parallel for
   for (int k=0; k<Knew; k++)
   {
#ifdef USEFULL
   	k2ztDishPtrs[k]->set_stats(fNArr[k], fcountsArr + (k)*hyper.getD());
#else
   	k2ztDishPtrs[k]->set_stats(fNArr[k], fcountsArr[k]);
#endif
   }

   return newloglikelihoodz;
}
void hdp_subclusters::splitmerge_find_bestNumTables(vector< unordered_map<int,restaurantDishComps> > &trestaurants,
	vector<dishComps*> &k2ztDishPtrs, vector<int>& mnew, vector<int>& mold)
{
   int oldBetaAvg = alpha*(1.0/K)*100 - 0.5;
   int newBetaAvg = alpha*(1.0/mnew.size())*100 - 0.5;
   if (oldBetaAvg>=100 || newBetaAvg>=100)
   	mexErrMsgTxt("bestNumTables not big enough\n");
   oldBetaAvg *= bestNumTablesN;
   newBetaAvg *= bestNumTablesN;

   // loop through each document and find the best table count
   reduction_array<int> moldArr(NThreads, K, 0);
   reduction_array<int> mnewArr(NThreads, mnew.size(), 0);
   #pragma omp parallel for
   for (int d=0; d<D; d++)
   {
   	int proc = omp_get_thread_num();
   	for (int k=0; k<K-1; k++)
   	{
   		int dishInd = k2ztDishPtrs[k]->dishInd;
   		int moldk = bestNumTables[restaurants[d][dishInd].getNumCustomers() + oldBetaAvg];
   		int mnewk = bestNumTables[trestaurants[d][dishInd].getNumCustomers() + newBetaAvg];
   		moldArr.reduce_add(proc, k, moldk);
   		mnewArr.reduce_add(proc, k, mnewk);
   	}
   	// K
   	int dishInd = k2zDishPtrs[K-1]->dishInd;
		int moldk = bestNumTables[restaurants[d][dishInd].getNumCustomers() + oldBetaAvg];
		moldArr.reduce_add(proc, K-1, moldk);

		// K & K+1
		if (mnew.size()>mold.size())
		{
			for (int k=K-1; k<K+1; k++)
			{
	   		int dishInd = k2ztDishPtrs[k]->dishInd;
	   		int mnewk = bestNumTables[trestaurants[d][dishInd].getNumCustomers() + newBetaAvg];
	   		mnewArr.reduce_add(proc, k, mnewk);
	   	}
	   }
   }

   // aggregate and set them properly
   int* fmoldArr = moldArr.final_reduce_add();
   int* fmnewArr = mnewArr.final_reduce_add();
   memcpy(mold.data(), fmoldArr, sizeof(int)*K);
   memcpy(mnew.data(), fmnewArr, sizeof(int)*mnew.size());
}


void hdp_subclusters::splitmerge_sample_betas_bestNumTables(vector<dishComps*> &k2ztDishPtrs,
	vector<int> &mnew, vector<int> &mold, int& mnewTotal, int& moldTotal)
{
	int tmnewTotal = 0;
   int tmoldTotal = 0;
   Real total = gsl_ran_gamma(rArr[0], gamma, 1);
   #pragma omp parallel for reduction(+:total,tmoldTotal,tmnewTotal)
   for (int k=0; k<K-1; k++)
   {
   	int proc = omp_get_thread_num();
   	Real beta = gsl_ran_gamma(rArr[proc], mnew[k], 1);
   	k2ztDishPtrs[k]->setBeta(beta);
   	total += beta;
   	tmoldTotal += mold[k];
   	tmnewTotal += mnew[k];
   }
   // K
	tmoldTotal += mold[K-1];
	// K:K+1
	if (mnew.size()>mold.size())
	{
		for (int k=K-1; k<K+1; k++)
		{
			k2ztDishPtrs[k]->setBeta(gsl_ran_gamma(rArr[0], mnew[k], 1));
		   tmnewTotal += mnew[k];
		   total += k2ztDishPtrs[k]->getBeta();
		}
	}

   moldTotal = tmoldTotal;
   mnewTotal = tmnewTotal;

   #pragma omp parallel for
   for (int k=0; k<k2ztDishPtrs.size(); k++)
   {
   	k2ztDishPtrs[k]->normalizeBeta(total);
   }
}


void hdp_subclusters::splitmerge_sample_temporary_params(vector< unordered_map<int,restaurantDishComps> > &trestaurants,
	vector<dishComps*> &k2ztDishPtrs)
{
	// the actual sampling is done in parallel here
	Real temptyBeta = 1;
	for (int k=0; k<k2ztDishPtrs.size(); k++)
	{
		k2ztDishPtrs[k]->sample();
		temptyBeta -= k2ztDishPtrs[k]->getBeta();
	}

	// sample the pi's
	#pragma omp parallel for
	for (int d=0; d<D; d++)
	{
		int proc = omp_get_thread_num();
		unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);
		Real total = gsl_ran_gamma(rArr[proc], alpha*temptyBeta, 1);
		/*if (total==mxGetInf())
		{
			mexPrintf("WTF?\n");
			mexPrintf("alpha=%e\n",alpha);
			mexPrintf("emptyBeta=%e\n", temptyBeta);
			mexErrMsgTxt("a\n");
		}*/
		for (int k=0; k<k2ztDishPtrs.size(); k++)
		{
			int dishInd = k2ztDishPtrs[k]->dishInd;
			total += (*trestaurantsd)[dishInd].sampleLogstick(rArr[proc], alpha, *(k2ztDishPtrs[k]));
		}
		total = fastlog(total);

		/*for (unordered_map<int,restaurantDishComps>::iterator rdishIt=trestaurantsd->begin(); rdishIt!=trestaurantsd->end(); rdishIt++)
		{
			Real temp = log(rdishIt->second.getStick()) - total;
			if (temp!=temp)
			{
				mexPrintf("%e\n", rdishIt->second.getStick());
				mexPrintf("%e\n", rdishIt->second.getLogstick());
				mexPrintf("NumCustomers=%d\n", rdishIt->second.getNumCustomers());
				mexPrintf("dishInd=%d\n", rdishIt->first);
				mexPrintf("beta=%e\n", tdishes[rdishIt->first].getBeta());
				mexPrintf("total = %e\n", total);
				mexErrMsgTxt("aa\n");
			}
		}*/

		for (unordered_map<int,restaurantDishComps>::iterator rdishIt=trestaurantsd->begin(); rdishIt!=trestaurantsd->end(); rdishIt++)
			rdishIt->second.normalizeLogstick(total);

		/*for (unordered_map<int,restaurantDishComps>::iterator rdishIt=trestaurantsd->begin(); rdishIt!=trestaurantsd->end(); rdishIt++)
			if (rdishIt->second.getLogstick()!=rdishIt->second.getLogstick())
			{
				mexPrintf("NumCustomers=%d\n", rdishIt->second.getNumCustomers());
				mexPrintf("dishInd=%d\n", rdishIt->first);
				mexPrintf("beta=%e\n", tdishes[rdishIt->first].getBeta());
				mexPrintf("total = %e\n", total);
				mexErrMsgTxt("a\n");
			}*/
	}
}








void hdp_subclusters::merge_create_temporary_vars(int k1, int k2, 
	vector< unordered_map<int,restaurantDishComps> > &trestaurants,
	unordered_map<int,dishComps> &tdishes, int &dishInd1, int &dishInd2)
{
	// find the two dishInds to merge
	dishInd1 = k2zDishPtrs[k1]->dishInd;
	dishInd2 = k2zDishPtrs[k2]->dishInd;

	//create the dishes
	/*for (int k=0; k<K; k++)
	{
		if (k!=k2)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			tdishes[dishInd] = dishComps(dishInd, 0, hyper, 0);
			tdishes[dishInd].param.copy(k2zDishPtrs[k]->main.param,true,false);
		}
	}
	tdishes[dishInd1].temp_merge_with(k2zDishPtrs[k2]->main);*/

	//create the restaurants / customers
	/*for (int d=0; d<D; d++)
	{
		// create the customer array
		unsigned int W = documents.getDocumentLength(d);
		tcustomers[d].resize(W);
		unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

		for (int k=0; k<K; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			restaurantDish* rdishPtr = &((*restaurantsd)[dishInd]);
			if (k==k1)
			{
				(*trestaurantsd)[dishInd] = rdishPtr->main;
				(*trestaurantsd)[dishInd].temp_merge_with((*restaurantsd)[dishInd2].main);
				(*trestaurantsd)[dishInd].clearNumCustomers();
			}
			else if (k!=k2)
			{
				(*trestaurantsd)[dishInd] = rdishPtr->main;
				(*trestaurantsd)[dishInd].clearNumCustomers();
			}
		}
	}*/




	//create the dishes
	dishComps temp(-1,0,hyper,0);
	for (int k=0; k<K; k++)
	{
		if (k!=k2)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			tdishes[dishInd] = dishComps(dishInd, 0, hyper, 0);
			//tdishes.insert(std::make_pair<int,dishComps>(dishInd, dishComps(dishInd, 0, hyper, 0)));
		}
	}

	#pragma omp parallel for
	for (int k=0; k<K; k++)
	{
		if (k!=k2)
		{
			// if k==k1, copy count so that we can resample the param
			int dishInd = k2zDishPtrs[k]->dishInd;
			tdishes[dishInd].param.copyNoNewMem(k2zDishPtrs[k]->main.param,k==k1);
			tdishes[dishInd].dishInd = dishInd;
		}
	}
	tdishes[dishInd1].temp_merge_with(k2zDishPtrs[k2]->main);

	//create the restaurants / customers
	#pragma omp parallel for
	for (int d=0; d<D; d++)
	{
		// create the customer array
		unsigned int W = documents.getDocumentLength(d);
		unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

		for (int k=0; k<K; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			restaurantDish* rdishPtr = &((*restaurantsd)[dishInd]);
			if (k==k1)
			{
				//(*trestaurantsd).insert(std::make_pair<int,restaurantDishComps>(dishInd, restaurantDishComps(rdishPtr->main)));
				(*trestaurantsd)[dishInd] = rdishPtr->main;
				(*trestaurantsd)[dishInd].temp_merge_with((*restaurantsd)[dishInd2].main);
				(*trestaurantsd)[dishInd].clearNumCustomers();
			}
			else if (k!=k2)
			{
				//(*trestaurantsd).insert(std::make_pair<int,restaurantDishComps>(dishInd, restaurantDishComps(rdishPtr->main)));
				(*trestaurantsd)[dishInd] = rdishPtr->main;
				(*trestaurantsd)[dishInd].clearNumCustomers();
			}
		}
	}



}

// find reverse z move probability
Real hdp_subclusters::merge_loglikelihood_zsplit(int k1, int k2,
	vector< unordered_map<int,restaurantDishComps> > &trestaurants,
   unordered_map<int,dishComps> &tdishes, vector<dishComps*> &k2ztDishPtrs)
{
	Real oldloglikelihoodz = 0;
	// sample a new theta
	dishComps* tsplitDish1 = &(k2zDishPtrs[k1]->main);
	dishComps* tsplitDish2 = &(k2zDishPtrs[k2]->main);
	int dishInd1 = tsplitDish1->dishInd;
	int dishInd2 = tsplitDish2->dishInd;

	//oldloglikelihoodz += k2zDishPtrs[k1]->main.param.getlogmu_posterior();
	//oldloglikelihoodz += k2zDishPtrs[k2]->main.param.getlogmu_posterior();

	// propose a new z~p(z|~theta,~pi)
	#pragma omp parallel for schedule(dynamic) reduction(+:oldloglikelihoodz)
   for (int d=0; d<D; d++)
   {
   	int proc = omp_get_thread_num();

      unsigned int* words = documents.getDocument(d);
      unsigned int W = documents.getDocumentLength(d);
      int* customersd = customers[d].data();
      unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);

      // this is how much it should be total
      Real logstickTotal = (*trestaurantsd)[dishInd1].getLogstick();
      // this is the relative proportions
      Real oldlogstick1 = restaurants[d][dishInd1].getLogstick();
      Real oldlogstick2 = restaurants[d][dishInd2].getLogstick();
      // stick1/(stick1+stick2) * stickTotal
      // log(stick1) - log(stick1 + stick2) + log(stickTotal)
      Real logstick1 = oldlogstick1 - logsumexp(oldlogstick1,oldlogstick2) + logstickTotal;
      Real logstick2 = oldlogstick2 - logsumexp(oldlogstick1,oldlogstick2) + logstickTotal;

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      {
      	int thisDishInd = customersd[w]/2;
      	unsigned int word = words[w];

      	// go through and calculate the loglikelihoods for each available topic
      	Real prob;
#ifdef USELOGLIKELIHOOD
      	Real totalProb = -mxGetInf();
      	for (int k=0; k<K; k++)
      	{
      		//mexPrintf("k=%d/%d\n", k,K);drawnow();
      		Real tprob;
      		int dishInd = k2zDishPtrs[k]->dishInd;
      		if (k==k1)
      			tprob = logstick1 + tsplitDish1->predictive_loglikelihood(word);
      		else if (k==k2)
      			tprob = logstick2 + tsplitDish2->predictive_loglikelihood(word);
      		else
	   			tprob = (*trestaurantsd)[dishInd].getLogstick() + k2ztDishPtrs[k]->predictive_loglikelihood(word);

	   		//tprob = restaurants[d][dishInd].getLogstick() + k2zDishPtrs[k]->predictive_loglikelihood(word);
				totalProb = logsumexp(totalProb, tprob);
   			if (dishInd==thisDishInd)
   				prob = tprob;
      	}
      	oldloglikelihoodz += prob - totalProb;
#else
      	mexErrMsgTxt("merge not implemented for likelihood (instead of loglikelihood)!\n");
#endif
	   }
   }
   return oldloglikelihoodz;
}

void hdp_subclusters::propose_merges3()
{
	if (rand()%1000>0)
		return;

	if (K<=1)
		return;

	int k1 = my_randi(rArr[0])%K;
	int k2 = my_randi(rArr[0])%(K-1);
	if (k2>=k1)
		k2++;
	else
	{
		// make sure r1 comes before r2
		int temp = k2;
		k2 = k1;
		k1 = temp;
	}

	if (!always_splittable && (!(k2zDishPtrs[k1]->splittable) || !(k2zDishPtrs[k2]->splittable)))
		return;

	// create the temporary dishes, restaurants, and customers
	vector< unordered_map<int,restaurantDishComps> > trestaurants(D);
	unordered_map<int,dishComps> tdishes;
	int dishInd1, dishInd2;
	//mexPrintf("creating\n");drawnow();
	merge_create_temporary_vars(k1,k2, trestaurants, tdishes, dishInd1, dishInd2);

	//mexPrintf("k2zt\n");drawnow();
	// create the k2ztDishPtrs
	// move the k2 to the end so that we can make them consistent
	dish* temp = k2zDishPtrs[K-1];
	k2zDishPtrs[K-1] = k2zDishPtrs[k2];
	k2zDishPtrs[k2] = temp;
	int tempk = z2k[K-1];
	z2k[K-1] = tempk;
	z2k[k2] = tempk;
	vector<dishComps*> k2ztDishPtrs(K-1);
	for (int k=0; k<K-1; k++)
	{
		int dishInd = k2zDishPtrs[k]->dishInd;
		k2ztDishPtrs[k] = &(tdishes[dishInd]);
	}
	k2 = K-1;

	//mexPrintf("ztilde\n");drawnow();
	// sample zhat~p(zhat | ~theta, ~pi)
	Real newloglikelihoodz = splitmerge_sample_ztilde(trestaurants, k2ztDishPtrs);
	//mexPrintf("ztilde theta\n");drawnow();
	newloglikelihoodz += k2ztDishPtrs[k1]->param.getlogmu_posterior();

	// find the "optimal" estimate of m, and sample betas
	//mexPrintf("bestnumtable\n");drawnow();
   vector<int> mnew(K-1,0);
   vector<int> mold(K,0);
   splitmerge_find_bestNumTables(trestaurants, k2ztDishPtrs, mnew, mold);

   // sample a beta from the m's
   //mexPrintf("sample beta\n");drawnow();
   int mnewTotal, moldTotal;
   splitmerge_sample_betas_bestNumTables(k2ztDishPtrs, mnew, mold, mnewTotal, moldTotal);

   // sample a set of parameters, pi,theta, from the posterior
   //mexPrintf("sample params\n");drawnow();
   splitmerge_sample_temporary_params(trestaurants, k2ztDishPtrs);

	// find the probability of making the reverse transition
	//mexPrintf("zsplit\n");drawnow();
	Real oldloglikelihoodz = merge_loglikelihood_zsplit(k1,k2, trestaurants, tdishes, k2ztDishPtrs);
	//mexPrintf("%e %e\n", newloglikelihoodz, oldloglikelihoodz);
	//mexErrMsgTxt("WAHOO\n");



	// ------------------------
   // Calculate the Hastings ratio!

   //mexPrintf("k=%d,%d \t", dishInd1, dishInd2); drawnow();
   Real HR = -loggamma + fastlog(999);

   HR += fastlog(1.0/(K+1)) - fastlog(1.0/(K*(K-1)/2));

   HR += gsl_sf_lngamma(gamma+moldTotal) - gsl_sf_lngamma(gamma+mnewTotal);
   //mexPrintf("%e\t", HR);drawnow();
   #pragma omp parallel for reduction(+:HR)
   for (int k=0; k<K-1; k++)
   {
   	int dishInd = k2zDishPtrs[k]->dishInd;
   	if (mold[k]>0) HR += mold[k]*fastlog(dishes[dishInd].getBeta()) - gsl_sf_lngamma(mold[k]);
   	if (mnew[k]>0) HR += gsl_sf_lngamma(mnew[k]) - mnew[k]*fastlog(tdishes[dishInd].getBeta());
   }
   if (mold[K-1]>0) HR += mold[K-1]*fastlog(dishes[dishInd2].getBeta()) - gsl_sf_lngamma(mold[K-1]);
   //mexPrintf("%e\t", HR);drawnow();
	#pragma omp parallel for reduction(+:HR)
	for (int d=0; d<D; d++)
	{
		for (int k=0; k<K-1; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			int oldNumCustomers = restaurants[d][dishInd].getNumCustomers();
			int newNumCustomers = trestaurants[d][dishInd].getNumCustomers();
			Real oldab = alpha*dishes[dishInd].getBeta();
			Real newab = alpha*tdishes[dishInd].getBeta();
			if (newab>0 && newNumCustomers>0) HR += gsl_sf_lngamma(newab + newNumCustomers) - gsl_sf_lngamma(newab);
			if (oldab>0 && oldNumCustomers>0) HR += gsl_sf_lngamma(oldab) - gsl_sf_lngamma(oldab + oldNumCustomers);
		}
		int oldNumCustomers = restaurants[d][dishInd2].getNumCustomers();
		Real oldab = alpha*dishes[dishInd2].getBeta();
		if (oldab>0 && oldNumCustomers>0) HR += gsl_sf_lngamma(oldab) - gsl_sf_lngamma(oldab + oldNumCustomers);
	}
	//mexPrintf("%e\t", HR);drawnow();
	//mexPrintf("%e\t", HR);
	// p(x|z)
	//#pragma omp parallel for reduction(+:HR)
	for (int k=0; k<K-1; k++)
	{
		int dishInd = k2zDishPtrs[k]->dishInd;
		HR += tdishes[dishInd].data_loglikelihood_marginalized() - dishes[dishInd].data_loglikelihood_marginalized();
	}
	HR -= dishes[dishInd2].data_loglikelihood_marginalized();
	//mexPrintf("%e\t", HR);drawnow();
	//mexPrintf("%e\t", HR);
	// q(z)
	HR += oldloglikelihoodz - newloglikelihoodz;
	//mexPrintf("%e\n", HR);drawnow();
	//mexPrintf("%e %e %e\n", oldloglikelihoodz, newloglikelihoodz, HR);drawnow();


	if ((HR>0 || my_rand(rArr[0]) < fastexp(HR)))
	{
      mexPrintf("MERGED!\n");

      // go through and fix the betas and z's
      // create the new dish
      vector<Real> tempBetaSubs(2, 0.5);

      dishes[dishInd1].merge_with(dishes[dishInd2]);
      dishes[dishInd1].setBeta(tdishes[dishInd1].getBeta());
      dishes[dishInd1].set_unsplittable();
      dishes[dishInd1].set_unmergeable();
      deleteDish(dishInd2);

      emptyBeta = 1;
      for (int k=0; k<K-1; k++)
      {
      	int dishInd = k2zDishPtrs[k]->dishInd;
      	k2zDishPtrs[k]->setBeta(tdishes[dishInd].getBeta());
      	emptyBeta -= k2zDishPtrs[k]->getBeta();
      }

      // now do the z's
      populate_k2z();

      #pragma omp parallel for
		for (int d=0; d<D; d++)
		{
			unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
			for (rdishIterator rdishIt=restaurantsd->begin(); rdishIt!=restaurantsd->end(); rdishIt++)
				rdishIt->second.clearNumCustomers();
		}

		#pragma omp parallel for
		for (int k=0; k<K; k++)
			k2zDishPtrs[k]->empty_params();

		// temporary reduction arrays
		reduction_array<int> NArr(NThreads, K*2, 0);
#ifdef USEFULL
		reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
		reduction_hash2<int> countsArr(NThreads, K*2);
#endif
		#pragma omp parallel for schedule(dynamic)
	   for (int d=0; d<D; d++)
	   {
	   	int proc = omp_get_thread_num();
	      unsigned int* words = documents.getDocument(d);
	      unsigned int W = documents.getDocumentLength(d);
	      int* customersd = customers[d].data();
	      int* tcustomersd = tcustomers[d].data();
	      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

	      // too hard to do random order of words... just do it in order
	      for (int w=0; w<W; w++)
	      {
	      	unsigned int word = words[w];

	      	// go through and calculate the loglikelihoods for each available topic
	      	int dishInd = tcustomersd[w]/2;
	      	int newk = z2k[dishInd];

				// left or right?
				Real logpsub, logpzsub;
				int subInd;
				if (dishInd==dishInd1)
					subInd = 0;
				else if (dishInd==dishInd2)
					subInd = 1;
				else
					subInd = (*restaurantsd)[dishInd].sampleSubDish(rArr[proc], dishes[dishInd], word, logpsub, logpzsub);
				int c = 2*dishInd + subInd;
				int bin = 2*newk + subInd;

				// accumulate stats
	   		customersd[w] = c;
	   		(*restaurantsd)[dishInd].addCustomer(c);

	   		NArr.reduce_inc(proc, bin);
	   		countsArr.reduce_inc(proc, bin, word);
		   }
	   }

	   arr(int) fNArr = NArr.final_reduce_add();
   	bool deleted = false;
#ifdef USEFULL
   	arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   	unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
	   #pragma omp parallel for reduction(||:deleted)
	   for (int k=0; k<K; k++)
	   {
	   	bool tempDeleted = true;
	   	int numCustomers = 0;
	   	for (int s=0; s<2; s++)
	   	{
	   		tempDeleted = tempDeleted && fNArr[2*k+s]==0;
	   		numCustomers += fNArr[2*k+s];
	   	}
	   	deleted = deleted || tempDeleted;
#ifdef USEFULL
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif
	   }

	   if (deleted)
	   {
	   	for (int k=0; k<K; k++)
	   		if (k2zDishPtrs[k]->getN()==0)
	   			deleteDish(k2zDishPtrs[k]->dishInd);
			populate_k2z();
	   }
   }
   else
   	populate_k2z();
}





// OPTION 3
void hdp_subclusters::split_create_temporary_vars(int splitk, 
	vector< unordered_map<int,restaurantDishComps> > &trestaurants,
	unordered_map<int,dishComps> &tdishes, int &oldDishInd, int &newDishInd)
{
	/*// find a new dishInd to split into
	oldDishInd = k2zDishPtrs[splitk]->dishInd;
	for (newDishInd=0; newDishInd<D; newDishInd++)
		if (dishes.count(newDishInd)==0)
			break;

	//create the dishes
	dishComps temp(-1,0,hyper,0);
	for (int k=0; k<K; k++)
	{
		int dishInd = k2zDishPtrs[k]->dishInd;
		if (k==splitk)
		{
			tdishes[oldDishInd] = temp;//dishComps(oldDishInd, 0, hyper, 0);
			tdishes[newDishInd] = temp;//dishComps(newDishInd, 0, hyper, 0);
			tdishes[oldDishInd].dishInd = oldDishInd;
			tdishes[newDishInd].dishInd = newDishInd;
			tdishes[oldDishInd].param.copy(k2zDishPtrs[k]->left.param,true,false);
			tdishes[newDishInd].param.copy(k2zDishPtrs[k]->right.param,true,false);
		}
		else
		{
			tdishes[dishInd] = temp;//dishComps(dishInd, 0, hyper, 0);
			tdishes[dishInd].dishInd = dishInd;
			tdishes[dishInd].param.copy(k2zDishPtrs[k]->main.param,true,false);
		}
	}*/

	/*//create the restaurants / customers
	for (int d=0; d<D; d++)
	{
		// create the customer array
		unsigned int W = documents.getDocumentLength(d);
		tcustomers[d].resize(W);
		unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

		for (int k=0; k<K; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			restaurantDish* rdishPtr = &((*restaurantsd)[dishInd]);
			if (k==splitk)
			{
				(*trestaurantsd)[oldDishInd] = rdishPtr->left;
				(*trestaurantsd)[newDishInd] = rdishPtr->right;
				Real newpil = rdishPtr->getStickL();
				Real newpir = rdishPtr->getStickR();
				Real oldpi = rdishPtr->getStick();
				(*trestaurantsd)[oldDishInd].setStick(oldpi*newpil);
				(*trestaurantsd)[newDishInd].setStick(oldpi*newpir);
				(*trestaurantsd)[oldDishInd].clearNumCustomers();
				(*trestaurantsd)[newDishInd].clearNumCustomers();
			}
			else
			{
				(*trestaurantsd)[dishInd] = rdishPtr->main;
				(*trestaurantsd)[dishInd].clearNumCustomers();
			}
		}
	}*/





	// find a new dishInd to split into
	oldDishInd = k2zDishPtrs[splitk]->dishInd;
	for (newDishInd=0; newDishInd<D; newDishInd++)
		if (dishes.count(newDishInd)==0)
			break;

	//create the dishes
	dishComps temp(-1,0,hyper,0);
	for (int k=0; k<K; k++)
	{
		tdishes[k2zDishPtrs[k]->dishInd] = temp;
		//tdishes.insert(std::make_pair<int,dishComps>(k2zDishPtrs[k]->dishInd, dishComps(temp)));
	}
	tdishes[newDishInd] = temp;
	#pragma omp parallel for
	for (int k=0; k<K; k++)
	{
		int dishInd = k2zDishPtrs[k]->dishInd;
		if (k==splitk)
		{
			tdishes[oldDishInd].dishInd = oldDishInd;
			tdishes[newDishInd].dishInd = newDishInd;
			tdishes[oldDishInd].param.copyNoNewMem(k2zDishPtrs[k]->left.param,false);
			tdishes[newDishInd].param.copyNoNewMem(k2zDishPtrs[k]->right.param,false);
		}
		else
		{
			tdishes[dishInd].dishInd = dishInd;
			tdishes[dishInd].param.copyNoNewMem(k2zDishPtrs[k]->main.param,false);
		}
	}

	//create the restaurants / customers
	#pragma omp parallel for
	for (int d=0; d<D; d++)
	{
		// create the customer array
		unsigned int W = documents.getDocumentLength(d);
		unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);
		unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

		for (int k=0; k<K; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			restaurantDish* rdishPtr = &((*restaurantsd)[dishInd]);
			if (k==splitk)
			{
				//(*trestaurantsd).insert(std::make_pair<int,restaurantDishComps>(oldDishInd, restaurantDishComps(rdishPtr->left)));
				//(*trestaurantsd).insert(std::make_pair<int,restaurantDishComps>(newDishInd, restaurantDishComps(rdishPtr->right)));
				(*trestaurantsd)[oldDishInd] = rdishPtr->left;
				(*trestaurantsd)[newDishInd] = rdishPtr->right;
				Real newpil = rdishPtr->getStickL();
				Real newpir = rdishPtr->getStickR();
				Real oldpi = rdishPtr->getStick();
				(*trestaurantsd)[oldDishInd].setStick(oldpi*newpil);
				(*trestaurantsd)[newDishInd].setStick(oldpi*newpir);
				(*trestaurantsd)[oldDishInd].clearNumCustomers();
				(*trestaurantsd)[newDishInd].clearNumCustomers();
			}
			else
			{
				//(*trestaurantsd).insert(std::make_pair<int,restaurantDishComps>(dishInd, restaurantDishComps(rdishPtr->main)));
				(*trestaurantsd)[dishInd] = rdishPtr->main;
				(*trestaurantsd)[dishInd].clearNumCustomers();
			}
		}
	}

}


// find reverse z move probability
Real hdp_subclusters::split_loglikelihood_zmerge(int splitk, int oldDishInd, int newDishInd,
	vector< unordered_map<int,restaurantDishComps> > &trestaurants,
   unordered_map<int,dishComps> &tdishes, vector<dishComps*> &k2ztDishPtrs)
{
	Real oldloglikelihoodz = 0;
	// sample a new theta
	dishComps tmergedDish(tdishes[oldDishInd]);
	tmergedDish.temp_merge_with(tdishes[newDishInd]);
	oldloglikelihoodz = tmergedDish.param.getlogmu_posterior();

	// propose a new z~p(z|~theta,~pi)
	#pragma omp parallel for schedule(dynamic) reduction(+:oldloglikelihoodz)
   for (int d=0; d<D; d++)
   {
   	int proc = omp_get_thread_num();

      unsigned int* words = documents.getDocument(d);
      unsigned int W = documents.getDocumentLength(d);
      int* customersd = customers[d].data();
      unordered_map<int,restaurantDishComps>* trestaurantsd = &(trestaurants[d]);

   	Real tmergedLogstick = logsumexp((*trestaurantsd)[oldDishInd].getLogstick(), (*trestaurantsd)[newDishInd].getLogstick());
   	Real tmergedStick = fastexp(tmergedLogstick);

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      {
      	int thisDishInd = customersd[w]/2;
      	unsigned int word = words[w];

      	// go through and calculate the loglikelihoods for each available topic
      	Real prob;
#ifdef USELOGLIKELIHOOD
      	Real totalProb = -mxGetInf();
      	for (int k=0; k<K; k++)
      	{
      		Real tprob;
      		int dishInd = k2ztDishPtrs[k]->dishInd;
      		if (dishInd==oldDishInd)
      			tprob = tmergedLogstick + tmergedDish.predictive_loglikelihood(word);
      		else
	   			tprob = (*trestaurantsd)[dishInd].getLogstick() + k2ztDishPtrs[k]->predictive_loglikelihood(word);
				totalProb = logsumexp(totalProb, tprob);
   			if (dishInd==thisDishInd)
   				prob = tprob;
      	}
      	oldloglikelihoodz += prob - totalProb;
#else
      	Real totalProb = 0;
      	for (int k=0; k<K; k++)
      	{
      		Real tprob;
      		int dishInd = k2zDishPtrs[k]->dishInd;
      		if (k==splitk)
      			tprob = tmergedStick * tmergedDish.predictive_likelihood(word);
      		else
	   			tprob = (*trestaurantsd)[dishInd].getStick() * (tdishes[dishInd].predictive_likelihood(word));
   			totalProb += tprob;
   			if (dishInd==thisDishInd)
   				prob = tprob;
      	}
      	oldloglikelihoodz += fastlog(prob/totalProb);
#endif
	   }
   }
   return oldloglikelihoodz;
}



void hdp_subclusters::propose_splits3()
{
	// propose to split cluster splitk
	int splitk = my_randi(rArr[0])%K;
	if (!always_splittable && !(k2zDishPtrs[splitk]->splittable))
		return;

	// create the temporary dishes, restaurants, and customers
	vector< unordered_map<int,restaurantDishComps> > trestaurants(D);
	unordered_map<int,dishComps> tdishes;
	int oldDishInd, newDishInd;
	split_create_temporary_vars(splitk, trestaurants, tdishes, oldDishInd, newDishInd);

	// create the k2ztDishPtrs
	vector<dishComps*> k2ztDishPtrs(K+1);
	for (int k=0; k<K; k++)
	{
		int dishInd = k2zDishPtrs[k]->dishInd;
		k2ztDishPtrs[k] = &(tdishes[dishInd]);
	}
	k2ztDishPtrs[K] = &(tdishes[newDishInd]);


	// sample zhat~p(zhat | ~theta, ~pi)
	Real newloglikelihoodz = splitmerge_sample_ztilde(trestaurants, k2ztDishPtrs);

	// find the "optimal" estimate of m, and sample betas
   vector<int> mnew(K+1,0);
   vector<int> mold(K,0);
   splitmerge_find_bestNumTables(trestaurants, k2ztDishPtrs, mnew, mold);

   // sample a beta from the m's
   int mnewTotal, moldTotal;
   splitmerge_sample_betas_bestNumTables(k2ztDishPtrs, mnew, mold, mnewTotal, moldTotal);

   // sample a set of parameters, pi,theta, from the posterior
   splitmerge_sample_temporary_params(trestaurants, k2ztDishPtrs);

	// find the probability of making the reverse transition
	Real oldloglikelihoodz = split_loglikelihood_zmerge(splitk, oldDishInd, newDishInd, trestaurants, tdishes, k2ztDishPtrs);
	//mexPrintf("%e %e\n", newloglikelihoodz, oldloglikelihoodz);
	//mexErrMsgTxt("WAHOO\n");


   

   // ------------------------
   // Calculate the Hastings ratio!

   //mexPrintf("k=%d %d %d\t", oldDishInd, moldTotal, mnewTotal); drawnow();
   Real HR = loggamma - fastlog(999);

   HR += fastlog(1.0/(K*(K+1)/2)) - fastlog(1.0/K);

   HR += gsl_sf_lngamma(gamma+moldTotal) - gsl_sf_lngamma(gamma+mnewTotal);
   //mexPrintf("%e\t", HR);
   #pragma omp parallel for reduction(+:HR)
   for (int k=0; k<K; k++)
   {
   	int dishInd = k2zDishPtrs[k]->dishInd;
   	if (mold[k]>0) HR += mold[k]*fastlog(dishes[dishInd].getBeta()) - gsl_sf_lngamma(mold[k]);
   	if (mnew[k]>0) HR += gsl_sf_lngamma(mnew[k]) - mnew[k]*fastlog(tdishes[dishInd].getBeta());
   }
   if (mnew[K]>0) HR += gsl_sf_lngamma(mnew[K]) - mnew[K]*fastlog(tdishes[newDishInd].getBeta());
   //mexPrintf("%e\t", HR);
	#pragma omp parallel for reduction(+:HR)
	for (int d=0; d<D; d++)
	{
		for (int k=0; k<K; k++)
		{
			int dishInd = k2zDishPtrs[k]->dishInd;
			int oldNumCustomers = restaurants[d][dishInd].getNumCustomers();
			int newNumCustomers = trestaurants[d][dishInd].getNumCustomers();
			Real oldab = alpha*dishes[dishInd].getBeta();
			Real newab = alpha*tdishes[dishInd].getBeta();
			if (newab>0 && newNumCustomers>0) HR += gsl_sf_lngamma(newab + newNumCustomers) - gsl_sf_lngamma(newab);
			if (oldab>0 && oldNumCustomers>0) HR += gsl_sf_lngamma(oldab) - gsl_sf_lngamma(oldab + oldNumCustomers);
		}
		int newNumCustomers = trestaurants[d][newDishInd].getNumCustomers();
		Real newab = alpha*tdishes[newDishInd].getBeta();
		if (newab>0 && newNumCustomers>0) HR += gsl_sf_lngamma(newab + newNumCustomers) - gsl_sf_lngamma(newab);
	}
	//mexPrintf("%e\t", HR);drawnow();
	//mexPrintf("%e\t", HR);
	// p(x|z)
	//#pragma omp parallel for reduction(+:HR)
	for (int k=0; k<K; k++)
	{
		int dishInd = k2zDishPtrs[k]->dishInd;
		HR += tdishes[dishInd].data_loglikelihood_marginalized() - dishes[dishInd].data_loglikelihood_marginalized();
	}
	HR += tdishes[newDishInd].data_loglikelihood_marginalized();
	//mexPrintf("%e\t", HR);drawnow();
	//mexPrintf("%e\t", HR);
	// q(z)
	HR += oldloglikelihoodz - newloglikelihoodz;
	//mexPrintf("%e\n", HR);drawnow();
	//mexPrintf("%e %e %e\n", oldloglikelihoodz, newloglikelihoodz, HR);drawnow();


	if ((HR>0 || my_rand(rArr[0]) < fastexp(HR)))
	{
      //mexPrintf("NumSplits=%d\n", 1);

      // go through and fix the betas and z's
      // create the new dish
      vector<Real> tempBetaSubs(2, 0.5);

      dishes[newDishInd] = dish(newDishInd, tdishes[newDishInd].getBeta(), tempBetaSubs.data(), hyper, -mxGetInf(), mxGetInf(), false,-mxGetInf(),mxGetInf(),false);
      dishes[oldDishInd] = dish(oldDishInd, tdishes[oldDishInd].getBeta(), tempBetaSubs.data(), hyper, -mxGetInf(), mxGetInf(), false,-mxGetInf(),mxGetInf(),false);
      ////dishes.erase(newDishInd);
      //dishes.erase(oldDishInd);
      //dishes.insert(std::make_pair<int,dish>(newDishInd, dish(newDishInd, tdishes[newDishInd].getBeta(), tempBetaSubs.data(), hyper, -mxGetInf(), mxGetInf(), false,-mxGetInf(),mxGetInf(),false)));
      //dishes.insert(std::make_pair<int,dish>(oldDishInd, dish(oldDishInd, tdishes[newDishInd].getBeta(), tempBetaSubs.data(), hyper, -mxGetInf(), mxGetInf(), false,-mxGetInf(),mxGetInf(),false)));


      for (int d=0; d<D; d++)
      {
      	//restaurants[d].insert(std::make_pair<int,restaurantDish>(newDishInd, restaurantDish(0)));
      	restaurants[d][newDishInd] = restaurantDish(0);
      }

      emptyBeta = 1 - dishes[oldDishInd].getBeta();
      for (int k=0; k<K; k++)
      {
      	int dishInd = k2zDishPtrs[k]->dishInd;
      	k2zDishPtrs[k]->setBeta(tdishes[dishInd].getBeta());
      	emptyBeta -= k2zDishPtrs[k]->getBeta();
      }

      // now do the z's
      populate_k2z();

      #pragma omp parallel for
		for (int d=0; d<D; d++)
		{
			unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
			for (rdishIterator rdishIt=restaurantsd->begin(); rdishIt!=restaurantsd->end(); rdishIt++)
				rdishIt->second.clearNumCustomers();
		}

		#pragma omp parallel for
		for (int k=0; k<K; k++)
			k2zDishPtrs[k]->empty_params();

		// temporary reduction arrays
		reduction_array<int> NArr(NThreads, K*2, 0);
#ifdef USEFULL
		reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
		reduction_hash2<int> countsArr(NThreads, K*2);
#endif
		#pragma omp parallel for schedule(dynamic)
	   for (int d=0; d<D; d++)
	   {
	   	int proc = omp_get_thread_num();
	      unsigned int* words = documents.getDocument(d);
	      unsigned int W = documents.getDocumentLength(d);
	      int* customersd = customers[d].data();
	      int* tcustomersd = tcustomers[d].data();
	      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

	      // too hard to do random order of words... just do it in order
	      for (int w=0; w<W; w++)
	      {
	      	unsigned int word = words[w];

	      	// go through and calculate the loglikelihoods for each available topic
	      	int dishInd = tcustomersd[w]/2;
	      	int newk = z2k[dishInd];

				// left or right?
				Real logpsub, logpzsub;
				int subInd;
				if (dishInd==oldDishInd || dishInd==newDishInd)
					subInd = my_randi(rArr[proc])%2;
				else
					subInd = (*restaurantsd)[dishInd].sampleSubDish(rArr[proc], dishes[dishInd], word, logpsub, logpzsub);
				int c = 2*dishInd + subInd;
				int bin = 2*newk + subInd;

				// accumulate stats
	   		customersd[w] = c;
	   		(*restaurantsd)[dishInd].addCustomer(c);

	   		NArr.reduce_inc(proc, bin);
	   		countsArr.reduce_inc(proc, bin, word);
		   }
	   }

	   arr(int) fNArr = NArr.final_reduce_add();
   	bool deleted = false;
#ifdef USEFULL
   	arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   	unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
	   #pragma omp parallel for reduction(||:deleted)
	   for (int k=0; k<K; k++)
	   {
	   	bool tempDeleted = true;
	   	int numCustomers = 0;
	   	for (int s=0; s<2; s++)
	   	{
	   		tempDeleted = tempDeleted && fNArr[2*k+s]==0;
	   		numCustomers += fNArr[2*k+s];
	   	}
	   	deleted = deleted || tempDeleted;
#ifdef USEFULL
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif
	   }

	   if (deleted)
	   {
	   	for (int k=0; k<K; k++)
	   		if (k2zDishPtrs[k]->getN()==0)
	   			deleteDish(k2zDishPtrs[k]->dishInd);
			populate_k2z();
	   }
   }
}































void hdp_subclusters::merger_mergeCustomersLoglikelihood(vector<int>& merge_proposals,
	int Ks, vector<Real>& loglikelihoods)
{
	for (int it=0; it<Ks/2; it++)
	{
		int k1 = merge_proposals[2*it];
		int k2 = merge_proposals[2*it+1];
		Real betaL = k2zDishPtrs[k1]->getBeta();
		Real betaR = k2zDishPtrs[k2]->getBeta();
		Real beta = betaL + betaR;
		Real ab = alpha*beta;
		Real abl = alpha*betaL;
		Real abr = alpha*betaR;

		int dishInd1 = k2zDishPtrs[k1]->dishInd;
		int dishInd2 = k2zDishPtrs[k2]->dishInd;
		Real loglikelihood = 0;
		#pragma omp parallel for reduction(+:loglikelihood)
		for (int d=0; d<D; d++)
		{
			int proc = omp_get_thread_num();
			int numCustomersL = restaurants[d][dishInd1].getNumCustomers();
			int numCustomersR = restaurants[d][dishInd2].getNumCustomers();
			int numCustomers = numCustomersL + numCustomersR;
			//if (numCustomers>0) loglikelihood += (gsl_sf_lngamma(ab + numCustomers) - gsl_sf_lngamma(ab));
			if (numCustomersL>0) loglikelihood += (gsl_sf_lngamma(abl) - gsl_sf_lngamma(abl + numCustomersL)) - (gsl_sf_lngamma(0.5*ab) - gsl_sf_lngamma(0.5*ab + numCustomersL));
			if (numCustomersR>0) loglikelihood += (gsl_sf_lngamma(abr) - gsl_sf_lngamma(abr + numCustomersR)) - (gsl_sf_lngamma(0.5*ab) - gsl_sf_lngamma(0.5*ab + numCustomersR));
			//if (numCustomersL>0) loglikelihood += (gsl_sf_lngamma(abl)) - (gsl_sf_lngamma(0.5*ab));
			//if (numCustomersR>0) loglikelihood += (gsl_sf_lngamma(abr)) - (gsl_sf_lngamma(0.5*ab));
		}
		loglikelihoods[it] = loglikelihood;
	}
}

// propose random splits/merges
void hdp_subclusters::propose_mergesr()
{
	//oKs /= 20;
	//oKs = max(oKs,1);
	int numMerges = 0;
	vector<int> merge_with(K,-1);
	vector<int> merge_proposals(K);
	for (int mergeit=0; mergeit<K; mergeit++)
	{
		int Ks = 0;
		for (int k=0; k<K; k++)
			if (merge_with[k]<0)
				merge_proposals[Ks++] = k;
		
		// do a Knuth shuffle
		for (int ks=0; ks<Ks; ks++)
		{
			int swapk = ks + my_randi(rArr[0]) % (Ks-ks);
			int temp = merge_proposals[ks];
			merge_proposals[ks] = merge_proposals[swapk];
			merge_proposals[swapk] = temp;
		}

		vector<int> moldL(Ks/2);
		vector<int> moldR(Ks/2);
		merge2_find_bestNumTables(merge_proposals, Ks, moldL, moldR);

		vector<Real> mergeCustomerLoglikelihoods(K);
	   merger_mergeCustomersLoglikelihood(merge_proposals, Ks, mergeCustomerLoglikelihoods);

	   //#pragma omp parallel for reduction(+:numMerges)
	   for (int it=0; it<Ks/2; it++)
	   {
	   	int proc = omp_get_thread_num();
	   	int km = merge_proposals[2*it];
	   	int kn = merge_proposals[2*it+1];
	   	int dishInd1 = k2zDishPtrs[km]->dishInd;
	   	int dishInd2 = k2zDishPtrs[kn]->dishInd;
	   	Real betaL = k2zDishPtrs[km]->getBeta();
	   	Real betaR = k2zDishPtrs[kn]->getBeta();
	   	Real beta = betaL + betaR;

	   	//mexPrintf("1(%d,%d)\t", dishInd1, dishInd2);
	   	Real HR = -loggamma - fastlog(100.0);

	   	Real temp = 2*(Ks/2);
	   	HR -= fastlog(K*temp / (Ks*(Ks-1)/2));

	   	HR += -gammalnint(moldL[it]) - gammalnint(moldR[it]) + gammalnint(moldL[it]+moldR[it]);
			HR += moldL[it]*fastlog(betaL) + moldR[it]*fastlog(betaR) - (moldL[it]+moldR[it])*fastlog(beta);
			//mexPrintf("%e\t", HR);

			HR += mergeCustomerLoglikelihoods[it];
			//mexPrintf("%e\t", HR);

			HR += k2zDishPtrs[km]->data_loglikelihood_marginalized_testmerge(*(k2zDishPtrs[kn]));
			HR += -k2zDishPtrs[km]->data_loglikelihood_marginalized() - k2zDishPtrs[kn]->data_loglikelihood_marginalized();
			//mexPrintf("%e\n", HR);

	      if ((HR>0 || my_rand(rArr[proc]) < fastexp(HR)))
	      {
	         merge_with[km] = km;
	         merge_with[kn] = km;
	         numMerges++;
	      }
	   }
	}

	vector<int> merge_counts(K,0);
	for (int k=0; k<K; k++)
		if (merge_with[k]>=0)
			merge_counts[merge_with[k]]++;
	for (int k=0; k<K; k++)
		if (merge_counts[k]!=0 && merge_counts[k]!=2)
			mexErrMsgTxt("uhoh\n");


   if (numMerges>0)
   {
   	//mexPrintf("NumMerges=%d\n", numMerges);
   	#pragma omp parallel for
		for (int d=0; d<D; d++)
		{
			unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);
			for (int k=0; k<K; k++)
				if (merge_with[k]>=0)
					(*restaurantsd)[k2zDishPtrs[k]->dishInd].clearNumCustomers();
		}

		#pragma omp parallel for
		for (int k=0; k<K; k++) if (merge_with[k]>=0)
			k2zDishPtrs[k]->empty_params();


		for (int kn=0; kn<K; kn++) if (merge_with[kn]>=0 && merge_with[kn]!=kn)
		{
			int dishInd2 = k2zDishPtrs[kn]->dishInd;
      	int km = merge_with[kn];
      	int dishInd1 = k2zDishPtrs[km]->dishInd;
         // merge the dish
         k2zDishPtrs[km]->merge_with(*(k2zDishPtrs[kn]));

         // merge the restaurant dishes
         #pragma omp parallel for
         for (int d=0; d<D; d++)
      		restaurants[d][dishInd1].merge_with(restaurants[d][dishInd2]);
      }


		reduction_array<int> NArr(NThreads, K*2, 0);
#ifdef USEFULL
	reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
	reduction_hash2<int> countsArr(NThreads, K*2);
#endif
		#pragma omp parallel for schedule(dynamic)
	   for (int d=0; d<D; d++)
	   {
	   	int proc = omp_get_thread_num();
	      unsigned int* words = documents.getDocument(d);
	      unsigned int W = documents.getDocumentLength(d);
	      int* customersd = customers[d].data();
	      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

	      // too hard to do random order of words... just do it in order
	      for (int w=0; w<W; w++)
	      {
	      	int k = z2k[customersd[w]/2];
	      	unsigned int word = words[w];

	      	if (merge_with[k]>=0)
				{
					int newk = merge_with[k];
					int dishInd = k2zDishPtrs[newk]->dishInd;

					int subInd = merge_with[k]!=k;
					int c = 2*dishInd + subInd;
					int bin = 2*newk + subInd;

					// accumulate stats
   				customersd[w] = c;
   				(*restaurantsd)[dishInd].addCustomer(c);

	   			NArr.reduce_inc(proc, bin);
   				countsArr.reduce_inc(proc, bin, word);
				}
	   	}
   	}

	   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   	arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   	unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif

	   #pragma omp parallel for
	   for (int k=0; k<K; k++) if (merge_with[k]>=0 && merge_with[k]==k)
	   {
	   	int numCustomers = 0;
   		for (int s=0; s<2; s++)
	   		numCustomers += fNArr[2*k+s];
#ifdef USEFULL
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif

   		k2zDishPtrs[k]->set_unsplittable();
   		k2zDishPtrs[k]->set_unmergeable();
   	}

   	for (int k=0; k<K; k++)
      	if (merge_with[k]>=0 && merge_with[k]!=k)
      		deleteDish(k2zDishPtrs[k]->dishInd);

      populate_k2z();
   }
}










void hdp_subclusters::construct_splitsr(vector<Real> &splitLoglikelihoods,
	vector<int> &mnewL, vector<int> &mnewR)
{
	// temporary reduction arrays
	reduction_array<int> NArr(NThreads, K*2, 0);
	reduction_array<Real> rsplitlikelihoodArr(NThreads, K, 0);
#ifdef USEFULL
	reduction_array2<int> countsArr(NThreads, K*2, hyper.getD(), 0);
#else
	reduction_hash2<int> countsArr(NThreads, K*2);
#endif

	// not really for reduction, just temporary space
	reduction_array<Real> pileftArr(NThreads, K, 0);

	int oldBetaAvg = alpha*(1.0/K)*100 - 0.5;
   int newBetaAvg = alpha*(1.0/(K+1))*100 - 0.5;
   if (oldBetaAvg>=100 || newBetaAvg>=100)
   	mexErrMsgTxt("bestNumTables not big enough\n");
   oldBetaAvg *= bestNumTablesN;
   newBetaAvg *= bestNumTablesN;

   reduction_array<int> mleft(NThreads, K, 0);
   reduction_array<int> mright(NThreads, K, 0);

	// propose a new z~p(z|~theta,~pi)
	#pragma omp parallel for schedule(dynamic)
   for (int d=0; d<D; d++)
   {
   	int proc = omp_get_thread_num();

   	arr(Real) tprobsArr = probsArr[proc];
   	arr(int) tprobsArrIndices = probsArrIndices[proc];

      unsigned int* words = documents.getDocument(d);
      unsigned int W = documents.getDocumentLength(d);
      int* customersd = customers[d].data();
      int* tcustomersd = tcustomers[d].data();
      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

      // generate the new left stick
      Real* pileft = pileftArr.getThreadArray(proc);
      for (int k=0; k<K; k++)
      {
      	Real beta = k2zDishPtrs[k]->getBeta();
      	Real pil = gsl_ran_gamma(rArr[proc], alpha*beta*0.5, 1);
      	Real pir = gsl_ran_gamma(rArr[proc], alpha*beta*0.5, 1);
      	Real total = pil + pir;
      	pileft[k] = pil / total;

      	// zero out the new left count
      	int dishInd = k2zDishPtrs[k]->dishInd;
      	(*restaurantsd)[dishInd].leftR = 0;
      }

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      {
      	unsigned int word = words[w];

      	int dishInd = customersd[w]/2;
      	int k = z2k[dishInd];
      	int subInd = (my_rand(rArr[proc]) < pileft[k]);
      	int c = 2*dishInd + subInd;
			int bin = 2*k + subInd;

			// accumulate stats
   		tcustomersd[w] = c;
   		if (subInd==0)
   			(*restaurantsd)[dishInd].leftR++;

   		NArr.reduce_inc(proc, bin);
   		countsArr.reduce_inc(proc, bin, word);
	   }

	   // update the likelihoods q(z)
	   for (int k=0; k<K; k++)
	   {
	   	Real loglikelihood = 0;
	   	int dishInd = k2zDishPtrs[k]->dishInd;
	   	int numCustomers = (*restaurantsd)[dishInd].getNumCustomers();
	   	int numCustomersL = (*restaurantsd)[dishInd].leftR;
	   	int numCustomersR = numCustomers - numCustomersL;
	   	Real beta = k2zDishPtrs[k]->getBeta();
	   	Real ab = alpha*beta;
	   	Real abl = ab*0.5;
	   	if (numCustomers>0)  loglikelihood += gsl_sf_lngamma(ab + numCustomers) - gsl_sf_lngamma(ab);
	   	if (numCustomersL>0) loglikelihood += gsl_sf_lngamma(abl) - gsl_sf_lngamma(abl + numCustomersL);
	   	if (numCustomersR>0) loglikelihood += gsl_sf_lngamma(abl) - gsl_sf_lngamma(abl + numCustomersR);
	   	rsplitlikelihoodArr.reduce_add(proc, k, loglikelihood);

	   	// find the numtables
	   	mleft.reduce_add(proc,k,bestNumTables[numCustomersL + newBetaAvg]);
			mright.reduce_add(proc,k,bestNumTables[numCustomersR + newBetaAvg]);
	   }
   }
   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif

   #pragma omp parallel for
   for (int k=0; k<K; k++)
   {
   	int proc = omp_get_thread_num();
   	Real loglikelihood = -mxGetInf();
   	if (fNArr[2*k]>0 && fNArr[2*k+1]>0)
		{
#ifdef USEFULL
   	loglikelihood = hyper.data_loglikelihood_marginalized_testmerge(fNArr[2*k], fcountsArr + 2*k*hyper.getD())
   		+ hyper.data_loglikelihood_marginalized_testmerge(fNArr[2*k+1], fcountsArr + (2*k+1)*hyper.getD());
#else
   	loglikelihood = hyper.data_loglikelihood_marginalized_testmerge(fNArr[2*k], fcountsArr[2*k])
   		+ hyper.data_loglikelihood_marginalized_testmerge(fNArr[2*k+1], fcountsArr[2*k+1]);
#endif
		}
   	loglikelihood += -k2zDishPtrs[k]->data_loglikelihood_marginalized();
   	rsplitlikelihoodArr.reduce_add(proc, k, loglikelihood);
   }

   arr(Real) frsplitlikelihoodArr = rsplitlikelihoodArr.final_reduce_add();
   splitLoglikelihoods.resize(K);
   memcpy(splitLoglikelihoods.data(), frsplitlikelihoodArr, sizeof(Real)*K);

   arr(int) fmleft = mleft.final_reduce_add();
   arr(int) fmright = mright.final_reduce_add();
   memcpy(mnewL.data(), fmleft, sizeof(int)*K);
   memcpy(mnewR.data(), fmright, sizeof(int)*K);
}
void hdp_subclusters::splitr_sample_betas_bestNumTables(vector<int> &mnewL, vector<int>& mnewR,
	vector<Real> &betaL, vector<Real> &betaR)
{
   #pragma omp parallel for
   for (int k=0; k<K; k++)
   {
   	int proc = omp_get_thread_num();
   	betaL[k] = gsl_ran_gamma(rArr[proc], mnewL[k], 1);
   	betaR[k] = gsl_ran_gamma(rArr[proc], mnewR[k], 1);
   	Real total = betaL[k] + betaR[k];
   	betaL[k] /= total;
   	betaR[k] /= total;
   }
}
void hdp_subclusters::splitr_splitCustomersLoglikelihood(const vector<Real> &newBetaL, const vector<Real> &newBetaR,
   vector<Real>& loglikelihoods)
{
	for (int k=0; k<K; k++)
	{
		Real beta = k2zDishPtrs[k]->getBeta();
		Real betaL = newBetaL[k]*beta;
		Real betaR = newBetaR[k]*beta;
		Real ab = alpha*beta;
		Real abl = alpha*betaL;
		Real abr = alpha*betaR;

		int dishInd = k2zDishPtrs[k]->dishInd;
		Real loglikelihood = 0;
		#pragma omp parallel for reduction(+:loglikelihood)
		for (int d=0; d<D; d++)
		{
			int proc = omp_get_thread_num();
			restaurantDish* rdishPtr = &(restaurants[d][dishInd]);
			int numCustomers = rdishPtr->getNumCustomers();
			int numCustomersL = rdishPtr->leftR;
			int numCustomersR = numCustomers - numCustomersL;

			if (numCustomers>0) loglikelihood += gsl_sf_lngamma(ab) - gsl_sf_lngamma(ab + numCustomers);
			if (numCustomersL>0) loglikelihood += gsl_sf_lngamma(abl + numCustomersL) - gsl_sf_lngamma(abl);
			if (numCustomersR>0) loglikelihood += gsl_sf_lngamma(abr + numCustomersR) - gsl_sf_lngamma(abr);
		}
		loglikelihoods[k] = loglikelihood;
	}
}
void hdp_subclusters::propose_splitsr()
{
	vector<Real> splitLoglikelihoods(K);
	vector<int> mnewL(K,0);
	vector<int> mnewR(K,0);
	// first find the optimal m count
	construct_splitsr(splitLoglikelihoods, mnewL, mnewR);


   // sample new betas for the split.
   vector<Real> newBetaL(K);
   vector<Real> newBetaR(K);
   splitr_sample_betas_bestNumTables(mnewL, mnewR, newBetaL, newBetaR);

   // priors
   vector<Real> splitCustomerLoglikelihoods(K);
   splitr_splitCustomersLoglikelihood(newBetaL, newBetaR, splitCustomerLoglikelihoods);

	vector<bool> do_split(K, false);
   int num_splits = 0;
   Real tempEmptyBeta = emptyBeta;
   #pragma omp parallel for reduction(+:num_splits)
   for (int k=0; k<K; k++)
   {
      // check to see if the subclusters have converged
      dish* dishPtr = k2zDishPtrs[k];
      int dishInd = dishPtr->dishInd;

      Real beta = dishPtr->getBeta();
      Real betaL = newBetaL[k] * beta;
      Real betaR = newBetaR[k] * beta;

		//mexPrintf("k=%d\t",dishInd);drawnow();
		Real HR = loggamma + fastlog(100.0);

		Real temp = 2*((K+1)/2);
		//HR += log(Ks*temp / (Ks*(Ks+1)/2));
		HR += fastlog(temp / (K*(K+1)/2));
		//mexPrintf("%e\t", HR);

		HR += gammalnint(mnewL[k]) + gammalnint(mnewR[k]) - gammalnint(mnewL[k]+mnewR[k]);
		HR += -mnewL[k]*fastlog(betaL) - mnewR[k]*fastlog(betaR) + (mnewL[k]+mnewR[k])*fastlog(beta);
		//mexPrintf("%e\t", HR);

		HR += splitLoglikelihoods[k] + splitCustomerLoglikelihoods[k];
		//mexPrintf("%e\n", HR);

      if ((HR>0 || my_rand(rArr[0]) < fastexp(HR)))
      {
         do_split[k] = true;
         num_splits++;
      }
   }

   if (num_splits>0)
   {
      //mexPrintf("NumSplits=%d\n", num_splits);
      int temp_num_splits = num_splits;
      // figure out which labels to split into and prepare simple split stuff
      vector<int> split_into(K);
      vector<int> split_into_k(K);

      int k = 0;
      for (k; k<K; k++)
         if (do_split[k])
            break;
      vector<Real> tempBetaSubs(2, 0.5);
      for (int nzh=0; nzh<D; nzh++)
      {
         if (dishes.count(nzh)==0)
         {
         	int dishInd1 = k2zDishPtrs[k]->dishInd;
            int dishInd2 = nzh;
            split_into[k] = nzh;
            split_into_k[k] = K+(num_splits-temp_num_splits);

            Real oldBeta = k2zDishPtrs[k]->getBeta();
            Real betaL = newBetaL[k] * oldBeta;
            Real betaR = newBetaR[k] * oldBeta;
            dishes[dishInd2] = dish(dishInd2, betaR, tempBetaSubs.data(), hyper,-mxGetInf(),mxGetInf(),false,-mxGetInf(),mxGetInf(),false);
				dishes[dishInd1].setBeta(betaL);

            // add the dish to all restaurants
            for (int d=0; d<D; d++)
            {
            	restaurants[d][dishInd1].clearNumCustomers();
            	restaurants[d][dishInd2] = restaurantDish(0);
            	restaurants[d][dishInd2].split_fromSub(1, restaurants[d][dishInd1]);
            	restaurants[d][dishInd1].split_fromSub(0, restaurants[d][dishInd1]);
            }

            // transfer over the theta's and beta's
            dishes[dishInd2].split_fromSub(1, *(k2zDishPtrs[k]));

         	// mark them as unsplittable
         	k2zDishPtrs[k]->set_unsplittable();
         	k2zDishPtrs[k]->set_unmergeable();

            // transfer over the theta's and beta's
	         dishes[dishInd1].split_fromSub(0, *(k2zDishPtrs[k]));

         	temp_num_splits--;
         	// do the next ones if there are any
            if (temp_num_splits>0)
            {
               for (k=k+1; k<K; k++)
                  if (do_split[k])
                     break;
            }
            else
               break;
         }
      }

      // correct the labels
      // temporary reduction array variables
      reduction_array<int> NArr(NThreads, (K+num_splits*(2-1))*2, 0);
#ifdef USEFULL
		reduction_array2<int> countsArr(NThreads, (K+num_splits)*2, hyper.getD(), 0);
#else
		reduction_hash2<int> countsArr(NThreads, (K+num_splits)*2);
#endif

		#pragma omp parallel for schedule(dynamic)
	   for (int d=0; d<D; d++)
	   {
	   	int proc = omp_get_thread_num();

	      unsigned int* words = documents.getDocument(d);
	      unsigned int W = documents.getDocumentLength(d);
	      int* customersd = customers[d].data();
	      int* tcustomersd = tcustomers[d].data();
	      unordered_map<int,restaurantDish>* restaurantsd = &(restaurants[d]);

	      for (int w=0; w<W; w++) if (do_split[z2k[customersd[w]/2]])
	      {
	      	int k = z2k[customersd[w]/2];
	      	unsigned int word = words[w];

	      	int s = tcustomersd[w]%2;
				int newk, dishInd;
	      	if (s==0)
	      	{
	      		dishInd = customersd[w]/2;
	      		newk = k;
	      	}
	      	else
	      	{
	      		dishInd = split_into[k];
	      		newk = split_into_k[k];
	      	}
				int isRight = my_rand(rArr[proc])<0.5;
				int newc = 2*dishInd + isRight;
	      	int bin = 2*newk + isRight;

	      	customersd[w] = newc;
	      	(*restaurantsd)[dishInd].addCustomer(newc);
	      	
	      	NArr.reduce_inc(proc, bin);
	      	countsArr.reduce_inc(proc, bin, word);
		   }
	   }

	   arr(int) fNArr = NArr.final_reduce_add();
#ifdef USEFULL
   	arr(int) fcountsArr = countsArr.final_reduce_add();
#else
   	unordered_map<int,int>* fcountsArr = countsArr.final_reduce_add();
#endif
	   #pragma omp parallel for
	   for (int k=0; k<K; k++) if (do_split[k])
	   {
	   	int dishInd1 = k2zDishPtrs[k]->dishInd;
#ifdef USEFULL
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k)*hyper.getD());
#else
   		k2zDishPtrs[k]->set_stats(fNArr + (2*k), fcountsArr + (2*k));
#endif
	   	k2zDishPtrs[k]->sample_subdishes();

	   	int dishInd2 = split_into[k];
	   	int newk = split_into_k[k];
#ifdef USEFULL
	   	dishes[dishInd2].set_stats(fNArr + (2*newk), fcountsArr + (2*newk)*hyper.getD());
#else
   		dishes[dishInd2].set_stats(fNArr + (2*newk), fcountsArr + (2*newk));
#endif
   		dishes[dishInd2].sample_subdishes();
	   	// need to sample 
	   	//	1) a new set of theta's for the subdishes
	   	// 2) a new set of pi's for all restaurants
	   }
	   populate_k2z();
	   sample_pis();
	   // sample counts is done next step anyway
   }
}













void hdp_subclusters::check()
{
	for (int d=0; d<D; d++)
	{
		unsigned int W = documents.getDocumentLength(d);
		int* customersd = customers[d].data();
		for (int w=0; w<W; w++)
      {
      	int dishInd = customersd[w]/2;
      	if (dishes.count(dishInd)==0)
      		mexErrMsgTxt("uhoh\n");
      }
	}

	/*mexPrintf("checking........\n");
	for (int k=0; k<K; k++)
	{
		int dishInd = k2zDishPtrs[k]->dishInd;
		int totalNumTables = 0;
		for (int d=0; d<D; d++)
			totalNumTables += restaurants[d][dishInd].getNumTables();
		mexPrintf("totalNumTables[%d]=%d  %d\n", dishInd, totalNumTables,k2zDishPtrs[k]->getNumTables());
		if (k2zDishPtrs[k]->getNumTables() != totalNumTables)
		{
			mexPrintf("k=%d\n", k);
			mexPrintf("dishInd=%d\n", dishInd);
			mexPrintf("sum(rest.tables)=%d\n", totalNumTables);
			mexPrintf("dish.tables=%d\n", k2zDishPtrs[k]->getNumTables());
			mexErrMsgTxt("WRONG TABLE COUNT\n");
		}
	}
	/*return;


	if (K!=dishes.size())
		mexErrMsgTxt("e1\n");

	for (int d=0; d<D; d++)
		if (restaurants[d].size()!=K)
			mexErrMsgTxt("e2\n");


	for (int d=0; d<D; d++)
	{
		unordered_map<int,restaurantDish> trestaurant;
      unsigned int W = documents.getDocumentLength(d);
      int* customersd = customers[d].data();

      // too hard to do random order of words... just do it in order
      for (int w=0; w<W; w++)
      {
      	int dishInd = customersd[w]/2;
      	trestaurant.emplace(dishInd,restaurantDish(0));
      	trestaurant.at(dishInd).main.numCustomers++;
      }

      if (trestaurant.size() > restaurants[d].size())
      {
      	mexPrintf("d=%d\n", d);
      	mexPrintf("K=%d\n", K);
      	mexPrintf("%d\n%d\n", trestaurant.size(), restaurants[d].size());
      	mexErrMsgTxt("Error0\n");
      }

      for (unordered_map<int,restaurantDish>::iterator rdishIt=trestaurant.begin(); rdishIt!=trestaurant.end(); rdishIt++)
      {
      	int dishInd = rdishIt->first;
      	if (restaurants[d].at(dishInd).main.numCustomers != rdishIt->second.main.numCustomers)
      	{
      		mexPrintf("d=%d\n", d);
      		mexPrintf("dishInd=%d\n", dishInd);
      		mexPrintf("%d\n%d\n", restaurants[d].at(dishInd).main.numCustomers, rdishIt->second.main.numCustomers);
      		mexErrMsgTxt("Error0.1\n");
      	}
      }
   }


   for (int k=0; k<K; k++)
   {
   	int dishInd = k2zDishPtrs[k]->dishInd;
   	int numCustomers = 0;
   	for (int d=0; d<D; d++)
   		numCustomers += restaurants[d][dishInd].getNumCustomers();
   	if (numCustomers != k2zDishPtrs[k]->getN())
   	{
   		mexPrintf("k=%d\n", k);
   		mexPrintf("%d\n", numCustomers);
   		mexPrintf("%d\n", k2zDishPtrs[k]->getN());
   		mexErrMsgTxt("Error1\n");
   	}
   }

/*	unordered_map<int,dish> tdishes;
	tdishes.emplace(-1, dish(0,hyper,0));

	for (int d=0; d<D; d++)
	{
		for (unordered_map<int,restaurantDish>::iterator rdishIt=restaurants[d].begin(); rdishIt!=restaurants[d].end(); rdishIt++)
		{
			int dishInd = rdishIt->first;
			if (dishes.count(dishInd)==0)
				mexErrMsgTxt("Error1\n");
			tdishes.emplace(dishInd, dish(0,hyper,0));
			tdishes[dishInd].numTables+=rdishIt->second.numTables;
		}
	}

	if (tdishes.size()!=dishes.size())
	{
		mexPrintf("tdishes.size()=%d\tdishes.size()=%d\n", tdishes.size(), dishes.size());
		mexErrMsgTxt("Error2\n");
	}
	for (unordered_map<int,dish>::iterator dishIt=dishes.begin(); dishIt!=dishes.end(); dishIt++)
	{
		int dishInd = dishIt->first;
		if (tdishes.at(dishInd).numTables != dishIt->second.numTables)
		{
			mexPrintf("tdishes[%d].numTables=%d\tdishes[%d].numTables=%d\n", dishInd, tdishes[dishInd].numTables, dishInd, dishIt->second.numTables);
			mexErrMsgTxt("Error3\n");
		}
	}*/
}
