// =============================================================================
// == calc_sdfIMPORT.cpp
// == --------------------------------------------------------------------------
// == The MEX interface file to calculate a signed distance function
// == --------------------------------------------------------------------------
// == Copyright 2011. MIT. All Rights Reserved.
// == Written by Jason Chang 06-13-2011
// =============================================================================

#define cache_line 4

#include "mex.h"
#include <iostream>
#include <math.h>
#include <algorithm>
#include <time.h>
#include "debugMEX.h"
#include "helperMEX.h"
#include "hdp_subclusters.h"
#include "dir_sampled.h"
#include "stopwatch.h"

#define NUMARGS 3
#define NUMOUT 6


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
   srand (time(NULL));
   // Check for proper number of arguments
   if (nrhs != NUMARGS) {
         mexErrMsgTxt("Incorrect number of input arguments required.");
   } else if (nlhs > NUMOUT) {
         mexErrMsgTxt("Too many output arguments expected.");
   }

   checkInput(prhs[0], STRUCT); //data
   checkInput(prhs[1], STRUCT); // params
   checkInput(prhs[2], STRUCT); // model

   const mxArray* data = prhs[0];
   const mxArray* params = prhs[1];

   int its_crf = getInput<double>(getField(params,0,"its_crf"));
   int NThreads = getInput<double>(getField(params,0,"NThreads"));

   omp_set_num_threads(NThreads);

   const mxArray* model = prhs[2];


   plhs[1] = mxCreateNumericMatrix(its_crf,1,mxDOUBLE_CLASS,mxREAL);
   plhs[2] = mxCreateNumericMatrix(its_crf,1,mxDOUBLE_CLASS,mxREAL);
   plhs[3] = mxCreateNumericMatrix(its_crf,1,mxDOUBLE_CLASS,mxREAL);
   plhs[4] = mxCreateNumericMatrix(its_crf,1,mxDOUBLE_CLASS,mxREAL);
   plhs[5] = mxCreateCellMatrix(its_crf,1);
   arr(double) timeArr = getArrayInput<double>(plhs[1]);
   arr(double) EArr = getArrayInput<double>(plhs[2]);
   arr(double) HOWArr = getArrayInput<double>(plhs[3]);
   arr(double) KnnzArr = getArrayInput<double>(plhs[4]);
   mxArray* nhistArr = plhs[5];



   //mexPrintf("Instantiation\n");drawnow();
   hdp_subclusters hdp(params, data);
   //mexPrintf("Initialization\n");drawnow();
   hdp.initialize(model);
   //mexPrintf("Done\n");drawnow();
   //hdp.check();

   vector<int> nhistvals;

   stopwatch timer;
   for (int it=0; it<its_crf; it++)
   {
      timer.tic();

      
      //mexPrintf("Sample Params\n");drawnow();
      hdp.sample_params();
      //hdp.check();


      //mexPrintf("Sample Dishes\n");drawnow();
      hdp.sample_dishes2();
      //hdp.sample_dishes3();
      //hdp.check();

      //hdp.propose_splits1();
      //mexPrintf("Propose Merges2\n");drawnow();
      hdp.propose_merges2();
      //mexPrintf("Propose Splits2\n");drawnow();
      hdp.propose_splits2();


      //mexPrintf("Propose Splits1\n");drawnow();
      /*hdp.propose_splits1();


      //mexPrintf("Random Merges\n");drawnow();
      hdp.propose_mergesr();
      //mexPrintf("Random Splits\n");drawnow();
      if (rand()%100==0)
         hdp.propose_splitsr();*/


      //mexPrintf("Propose Merges3\n");drawnow();
      hdp.propose_merges3();
      //mexPrintf("Propose Splits3\n");drawnow();
      hdp.propose_splits3();
      
      //mexPrintf("Sample Counts\n");drawnow();
      hdp.sample_counts();
      //hdp.check();

      //mexPrintf("Sample Betas\n");drawnow();
      hdp.sample_betas();
      //hdp.check();
      //hdp.sample_hyperparameters();

      double time = timer.toc();
      timeArr[it] = time;
      //mexPrintf("Calculate Likelihood\n");drawnow();
      EArr[it] = hdp.joint_loglikelihood();
      //mexPrintf("Calculate HOW Likelihood\n");drawnow();
      HOWArr[it] = hdp.HOW_loglikelihood();
      //mexPrintf("Done\n");drawnow();
      //hdp.check();
      KnnzArr[it] = hdp.getK();

      hdp.getnhist(nhistvals);
      mxSetCell(nhistArr, it, mxCreateArrayINT32<int>(nhistvals.size(),1, nhistvals.data()));
   }
   //mexPrintf("Output\n");drawnow();
   hdp.write_output(plhs[0]);
   //mexPrintf("Done\n");drawnow();
}
