#include "sample_categorical.h"
#include "myfuncs.h"

struct restaurantDishComps
{
   int numCustomers;
   int numTables;
   Real stick;
   Real logstick;
   restaurantDishComps() {}
   restaurantDishComps(int _numCustomers, int _numTables, Real _logstick) :
      numCustomers(_numCustomers), numTables(_numTables), logstick(_logstick) {}
   int getNumCustomers() const;
   Real getStick() const;
   Real getLogstick() const;
   Real setStick(Real _stick);
   Real setLogstick(Real _logstick);
   void addCustomer();
   void clearNumCustomers();
   Real sampleLogstick(gsl_rng* r, Real alpha, dishComps& theDish);
   void normalizeLogstick(Real total);
   void temp_merge_with(restaurantDishComps& other);
};
inline int restaurantDishComps::getNumCustomers() const        { return numCustomers; }
inline Real restaurantDishComps::getStick() const            { return stick; }
inline Real restaurantDishComps::getLogstick() const         { return logstick; }
inline Real restaurantDishComps::setStick(Real _stick)
{
   stick = _stick;
   logstick = log(stick);
}
inline Real restaurantDishComps::setLogstick(Real _logstick)
{
   logstick = _logstick;
   stick = exp(logstick);
}
inline void restaurantDishComps::addCustomer()
{
   ++numCustomers;
}
inline void restaurantDishComps::clearNumCustomers()
{
   numCustomers = 0;
}
inline Real restaurantDishComps::sampleLogstick(gsl_rng* r, Real alpha, dishComps& theDish)
{
   Real beta = theDish.getBeta();
   stick = gsl_ran_gamma(r, alpha*beta + numCustomers, 1);
   return stick;
}
inline void restaurantDishComps::normalizeLogstick(Real total)
{
   logstick = log(stick);
   if (logstick==-mxGetInf())
      stick = 0;
   else
      setLogstick(logstick - total);
}
inline void restaurantDishComps::temp_merge_with(restaurantDishComps& other)
{
   numCustomers += other.numCustomers;
   numTables += other.numTables;
   stick += other.stick;
   logstick = logsumexp(logstick, other.logstick);
}






struct restaurantDish
{
   restaurantDishComps main;
   restaurantDishComps left;
   restaurantDishComps right;

   int leftR;

   restaurantDish() {}
   restaurantDish(int numTables);
   restaurantDish(int numTables, int* numTablesSubs);

   void addCustomer(int c);

   int getNumCustomers() const;
   int getNumCustomersL() const;
   int getNumCustomersR() const;
   int getNumTables() const;
   int getNumTablesL() const;
   int getNumTablesR() const;

   //int getNumCustomersSub(int s);
   //int getNumTablesSub(int s);

   Real getStick() const;
   Real getStickL() const;
   Real getStickR() const;
   Real getLogstick() const;
   Real getLogstickL() const;
   Real getLogstickR() const;

   Real getLogstickMarginalized(Real alpha, dish& theDish);

   Real sampleLogstick(gsl_rng* r, Real alpha, dish& theDish);
   void normalizeLogstick(Real total);

   int sampleSubDish(gsl_rng* r, dish& theDish, int word, Real &logpsub, Real &logpzsub) const;
   Real calcSubDish(int c, dish& theDish, int word);

   void clearNumCustomers();
   void setNumCustomers(int* numCustomersIn);

   void setNumTables(int numTables);

   int sampleNumTables(gsl_rng* r, Real ab, arr(double) logstirling, int logstirlingW);
   int sampleNumTablesL(gsl_rng* r, Real ab, arr(double) logstirling, int logstirlingW);
   int sampleNumTablesR(gsl_rng* r, Real ab, arr(double) logstirling, int logstirlingW);

   //int sampleNumTablesSub(int s, gsl_rng* r, Real ab, arr(Real) logstirling, int logstirlingW);
   int sampleNumTables(gsl_rng* r, restaurantDishComps &rdish, Real ab, arr(double) logstirling, int logstirlingW);
   int sampleNumTables(gsl_rng* r, int numCustomers, Real ab, arr(double) logstirling, int logstirlingW) const;

   void split_fromSub(int s, restaurantDish& other);
   void merge_with(restaurantDish& other);
};


/*inline restaurantDish::restaurantDish() : numCustomers(0), numTables(0)
{
}*/

inline restaurantDish::restaurantDish(int numTables) :
   main(0,numTables,0), left(0,0,0), right(0,0,0)
{
}

inline restaurantDish::restaurantDish(int numTables, int* numTablesSubs) :
   main(0,numTables,0), left(0,numTablesSubs[0],0), right(0,numTablesSubs[1],0)
{
}

inline void restaurantDish::addCustomer(int c)
{
   main.numCustomers++;
   left.numCustomers += (c%2==0);
   right.numCustomers += (c%2);
}

inline int restaurantDish::getNumCustomers() const             { return main.numCustomers; }
inline int restaurantDish::getNumCustomersL() const            { return left.numCustomers; }
inline int restaurantDish::getNumCustomersR() const            { return right.numCustomers; }
inline int restaurantDish::getNumTables() const                { return main.numTables; }
inline int restaurantDish::getNumTablesL() const               { return left.numTables; }
inline int restaurantDish::getNumTablesR() const               { return right.numTables; }

//inline int restaurantDish::getNumCustomersSub(int s) { return (s==0) ? left.numCustomers : right.numCustomers; }
//inline int restaurantDish::getNumTablesSub(int s) { return (s==0) ? left.numTables : right.numTables; }

inline Real restaurantDish::getStick()  const       { return main.stick; }
inline Real restaurantDish::getStickL()  const       { return left.stick; }
inline Real restaurantDish::getStickR()  const       { return right.stick; }
inline Real restaurantDish::getLogstick()  const       { return main.logstick; }
inline Real restaurantDish::getLogstickL()  const       { return left.logstick; }
inline Real restaurantDish::getLogstickR()  const       { return right.logstick; }

inline Real restaurantDish::getLogstickMarginalized(Real alpha, dish& theDish)
{
   Real beta = theDish.getBeta();
   return log(alpha*beta + main.numCustomers);
}

inline Real restaurantDish::sampleLogstick(gsl_rng* r, Real alpha, dish& theDish)
{
   Real beta = theDish.getBeta();
   main.stick = gsl_ran_gamma(r, alpha*beta + main.numCustomers, 1);

   Real stickl = gsl_ran_gamma(r, alpha*theDish.getBetaL() + left.numCustomers, 1);
   Real stickr = gsl_ran_gamma(r, alpha*theDish.getBetaR() + right.numCustomers, 1);
   Real total = stickl + stickr;

   left.setStick(stickl / total);
   right.setStick(stickr / total);

   return main.stick;
}

inline void restaurantDish::normalizeLogstick(Real total)
{
   main.logstick = log(main.stick);
   if (main.logstick==-mxGetInf())
      main.stick = 0;
   else
      main.setLogstick(main.logstick - total);
}


#ifdef USELOGLIKELIHOOD
inline int restaurantDish::sampleSubDish(gsl_rng* r, dish& theDish, int word, Real &logpsub, Real &logpzsub) const
{
   Real probl = left.logstick + theDish.predictive_loglikelihoodL(word);
   Real probr = right.logstick + theDish.predictive_loglikelihoodR(word);
   Real total = logsumexp(probl,probr);

   int newsub = (my_rand(r) >= exp(probl - total));
   if (newsub==0)
   {
      logpsub = probl;
      logpzsub = probl - total;
   }
   else
   {
      logpsub = probr;
      logpzsub = probr - total;
   }
   return newsub;
}
inline Real restaurantDish::calcSubDish(int c, dish& theDish, int word)
{
   Real probl = left.logstick + theDish.predictive_loglikelihoodL(word);
   Real probr = right.logstick + theDish.predictive_loglikelihoodR(word);
   Real total = logsumexp(probl,probr);

   if (c%2==0)
      return probl - total;
   else
      return probr - total;
}
#else
inline int restaurantDish::sampleSubDish(gsl_rng* r, dish& theDish, int word, Real &logpsub, Real &logpzsub) const
{
   Real probl = left.stick * theDish.predictive_likelihoodL(word);
   Real probr = right.stick * theDish.predictive_likelihoodR(word);
   Real total = probl + probr;

   int newsub = (my_rand(r) >= probl/total);
   if (newsub==0)
   {
      logpsub = log(probl);
      logpzsub = log(probl / total);
   }
   else
   {
      logpsub = log(probr);
      logpzsub = log(probr / total);
   }
   return newsub;
}
inline Real restaurantDish::calcSubDish(int c, dish& theDish, int word)
{
   Real probl = left.stick * theDish.predictive_likelihoodL(word);
   Real probr = right.stick * theDish.predictive_likelihoodR(word);
   Real total = probl + probr;

   if (c%2==0)
      return log(probl / total);
   else
      return log(probr / total);
}
#endif


inline void restaurantDish::clearNumCustomers()
{
   main.numCustomers = 0;
   left.numCustomers = 0;
   right.numCustomers = 0;
}
inline void restaurantDish::setNumCustomers(int* numCustomersIn)
{
   left.numCustomers = numCustomersIn[0];
   right.numCustomers = numCustomersIn[1];
   main.numCustomers = left.numCustomers + right.numCustomers;
}

inline void restaurantDish::setNumTables(int numTables)
{
   main.numTables = numTables;
}

inline int restaurantDish::sampleNumTables(gsl_rng* r, Real ab, arr(double) logstirling, int logstirlingW)
{
   return sampleNumTables(r, main, ab, logstirling, logstirlingW);
}
inline int restaurantDish::sampleNumTablesL(gsl_rng* r, Real ab, arr(double) logstirling, int logstirlingW)
{
   return sampleNumTables(r, left, ab, logstirling, logstirlingW);
}
inline int restaurantDish::sampleNumTablesR(gsl_rng* r, Real ab, arr(double) logstirling, int logstirlingW)
{
   return sampleNumTables(r, right, ab, logstirling, logstirlingW);
}

/*inline int restaurantDish::sampleNumTablesSub(int s, gsl_rng* r, Real ab, arr(Real) logstirling, int logstirlingW)
{
   if (s==0)
      return sampleNumTables(r, left, ab, logstirling, logstirlingW);
   else
      return sampleNumTables(r, right, ab, logstirling, logstirlingW);
}*/

inline int restaurantDish::sampleNumTables(gsl_rng* r, restaurantDishComps &rdish, Real ab, arr(double) logstirling, int logstirlingW)
{
   int numTables = sampleNumTables(r, rdish.numCustomers, ab, logstirling, logstirlingW);
   rdish.numTables = numTables;
   return numTables;
}

inline int restaurantDish::sampleNumTables(gsl_rng* r, int numCustomers, Real ab, arr(double) logstirling, int logstirlingW) const
{
   if (numCustomers==0)
      return 0;

   Real logab = log(ab);
   Real randnum = my_rand(r);
   randnum = log(randnum) + gsl_sf_lngamma(ab + numCustomers) - gsl_sf_lngamma(ab);

   Real total = logstirling[numCustomers];
   int numTables = 0;
   while (randnum>total && numTables<logstirlingW-1)
   {
      ++numTables;
      /*if (numCustomers+numTables*logstirlingW<0 || numCustomers+numTables*logstirlingW >= logstirlingW*logstirlingW)
      {
         mexPrintf("numCustomers=%d\n", numCustomers);
         mexPrintf("numCustomers+numTables*logstirlingW=%d\n", numCustomers+numTables*logstirlingW);
         mexPrintf("logstirlingW=%d\n", logstirlingW);
         mexPrintf("numTables=%d\n", numTables);
         mexErrMsgTxt("uhoh2\n");
      }*/
      total = logsumexp(total, logstirling[numCustomers+numTables*logstirlingW] + numTables*logab);
   }
   return numTables;
}





inline void restaurantDish::split_fromSub(int s, restaurantDish& other)
{
   // transfer the betas
   main.numCustomers = 0;
   left.numCustomers = 0;
   right.numCustomers = 0;
   if (s==0)
      main.numTables = other.left.numTables;
   else
      main.numTables = other.right.numTables;
}

inline void restaurantDish::merge_with(restaurantDish& other)
{
   Real total = logsumexp(main.logstick, other.main.logstick);
   main.numCustomers += other.main.numCustomers;
   main.numTables += other.main.numTables;
   main.logstick = total;
}